
Imports System.Collections.Generic
Imports System.Collections.Specialized

Public Class SyntaxRichTextBox
    Inherits System.Windows.Forms.RichTextBox

    Friend TextModified As Boolean = False
    Friend TextMustBeSaved As Boolean = False
    Friend TextMustBeIndented As Boolean = False


    ' ===============================================================================================
    '  AUTOMATION LANGUAGE COLORS
    ' ===============================================================================================
    Private Color_Text As Color = Color.FromArgb(0, 0, 220)
    Private Color_KeywordText As Color = Color.FromArgb(80, 0, 0)
    Private Color_CommentText As Color = Color.FromArgb(0, 50, 50)
    Private Color_CommentBack As Color = Color.FromArgb(240, 250, 220)
    Private Color_BreakpointText As Color = Color.FromArgb(50, 0, 0)
    Private Color_BreakpointBack As Color = Color.FromArgb(255, 240, 100)

    Private Color_ActiveBack As Color = Color.FromArgb(255, 255, 190)
    Private Color_WarningBack As Color = Color.FromArgb(255, 220, 120)
    Private Color_NotFoundBack As Color = Color.FromArgb(255, 200, 170)
    Private Color_ErrorBack As Color = Color.FromArgb(255, 180, 150)


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  TOKENIZER
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Enum TokenizerStatus As Int32
        Normal = 0
        InsideToken = 1
        InsideString = 2
    End Enum
    Friend Function Tokenize(ByVal line As String) As String()
        Dim status As TokenizerStatus
        Dim sa(-1) As String
        For i As Int32 = 0 To line.Length - 1
            Select Case status
                Case TokenizerStatus.Normal
                    Select Case line(i)
                        Case " "c, Chr(9), "("c, ")"c
                        Case "'"c
                            Exit For ' comments are completely skipped 
                        Case """"c
                            status = TokenizerStatus.InsideString
                            ReDim Preserve sa(sa.Length)
                        Case Else
                            status = TokenizerStatus.InsideToken
                            ReDim Preserve sa(sa.Length)
                            sa(sa.Length - 1) += line(i)
                    End Select
                Case TokenizerStatus.InsideToken
                    Select Case line(i)
                        Case " "c, Chr(9), "("c, ")"c
                            ' ------------------------------------------ end of token - eval right side ?  
                            status = TokenizerStatus.Normal
                            Select Case sa.Length
                                Case 1
                                    Select Case sa(0).ToLower
                                        Case "load", "save", "print", "speed", "presskeys", "sendkeys"
                                            ' -------------------------- isolate right side formula (trimmed) and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1)
                                            RemoveComments(sa(1))
                                            sa(1) = sa(1).Trim
                                            ' ------------------------------------------ stringtofile
                                            If sa(1).ToLower.StartsWith("stringtofile") Then
                                                ReDim Preserve sa(3)
                                                Dim p2 As Int32 = InStrRev(sa(1), " ")
                                                If p2 > 0 Then
                                                    sa(3) = sa(1).Substring(p2)
                                                    sa(2) = sa(1).Substring(12, p2 - 12)
                                                    sa(1) = "stringtofile"
                                                End If
                                            End If
                                            ' ------------------------------------------- deletefile
                                            If sa(1).ToLower.StartsWith("deletefile") Then
                                                ReDim Preserve sa(2)
                                                sa(2) = sa(1).Substring(10)
                                                sa(1) = "deletefile"
                                            End If
                                            ' -------------------------------------------- exit for
                                            Exit For
                                        Case "label"
                                            ' -------------------------- isolate right side formula (without "") and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1).Trim.Replace("""", "")
                                            RemoveComments(sa(1))
                                            Exit For
                                        Case "goto", "gosub"
                                            ' -------------------------- isolate right side formula and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1).Trim
                                            RemoveComments(sa(1))
                                            Exit For
                                        Case "if", "select", "case"
                                            ' -------------------------- isolate right side formula and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1).Trim
                                            RemoveComments(sa(1))
                                            Exit For
                                        Case "slot"
                                            ' -------------------------- isolate left side formula and continue
                                            Dim eq As Int32 = InStr(line, "=")
                                            If eq > i Then
                                                ReDim Preserve sa(1)
                                                sa(1) = line.Substring(i, eq - i - 2)
                                                i = eq - 2
                                            End If
                                    End Select
                                Case 2
                                    Select Case sa(0).ToLower
                                        Case "wait"
                                            If sa(1).ToLower <> "seconds" And _
                                               sa(1).ToLower <> "button" And _
                                               sa(1).ToLower <> "application" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(1)
                                                sa(1) = line.Substring(5).Trim
                                                RemoveComments(sa(1))
                                                Exit For
                                            End If
                                        Case "com"
                                            If sa(1).ToLower = "open" Then
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, sa(1).ToLower).Trim
                                                RemoveComments(sa(2))
                                                Dim n() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(2)).Split(" "c)
                                                If n.Length = 2 Then
                                                    ReDim Preserve sa(3)
                                                    sa(2) = n(0)
                                                    sa(3) = n(1)
                                                    Exit For
                                                End If
                                            ElseIf sa(1).ToLower <> "writestring" Or sa(1).ToLower <> "writeline" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, sa(1).ToLower).Trim
                                                RemoveComments(sa(2))
                                                Exit For
                                            End If
                                        Case Else
                                            If IsSomeVariable(sa(0)) Then
                                                If sa(1) = "=" Then
                                                    Dim pos As Int32 = line.IndexOf("=") + 1
                                                    If line.Substring(pos).Trim.ToLower.StartsWith("input") Then
                                                        ' -------------------------- isolate input prompt
                                                        ReDim Preserve sa(3)
                                                        sa(2) = "input"
                                                        sa(3) = GetRightPartAfterThisLowerCaseWord(line, "input")
                                                        RemoveComments(sa(3))
                                                        Exit For
                                                    Else
                                                        ' -------------------------- isolate right side formula
                                                        ReDim Preserve sa(2)
                                                        sa(2) = line.Substring(pos).Trim
                                                        RemoveComments(sa(2))
                                                    End If
                                                    ' ----------------------------- and exit
                                                    Exit For
                                                End If
                                            End If
                                    End Select
                                Case 3
                                    Select Case sa(0).ToLower
                                        Case "slot"
                                            If sa(2) = "=" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = line.Substring(i + 1)
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                        Case "wait"
                                            If sa(1).ToLower = "seconds" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, "seconds").Trim
                                                RemoveComments(sa(2))
                                                Exit For
                                            End If
                                            If sa(1).ToLower = "application" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, "application").Trim
                                                RemoveComments(sa(2))
                                                Exit For
                                            End If
                                        Case "button"
                                            If sa(2).ToLower = "text" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = GetRightPartAfterThisLowerCaseWord(line, "text").Trim
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                            If sa(2).ToLower = "color" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = GetRightPartAfterThisLowerCaseWord(line, "color").Trim
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                        Case "key"
                                            If sa(2).ToLower = "goto" OrElse sa(2).ToLower = "gosub" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = GetRightPartAfterThisLowerCaseWord(line, sa(2).ToLower).Trim
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                    End Select
                            End Select
                        Case "'"c
                            Exit For ' comments are completely skipped 
                        Case Else
                            sa(sa.Length - 1) += line(i)
                    End Select
                Case TokenizerStatus.InsideString
                    Select Case line(i)
                        Case """"c
                            If i >= line.Length - 2 OrElse line(i + 1) <> """"c Then
                                status = TokenizerStatus.Normal
                            Else
                                sa(sa.Length - 1) += line(i)
                                i += 1
                            End If
                        Case Else
                            sa(sa.Length - 1) += line(i)
                    End Select
            End Select
        Next
        Return sa
    End Function


    ' ===============================================================================================
    '  DISABLE MOUSE SELECTION WHEN NOT EDITING
    ' ===============================================================================================
    'Private Sub RTB_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseLeave
    '    If Not Running Then Enabled = True
    'End Sub
    'Private Sub RTB_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
    '    If Not Running Then Enabled = False
    'End Sub

    ' ===============================================================================================
    '  ENABLE DISABLE PAINT and DISABLE ZOOM with CTRL WHEEL
    ' ===============================================================================================
    Private Const MK_CONTROL As Int32 = &H8
    Private Const WM_MOUSEWHEEL As Int32 = &H20A
    Friend PaintEventEnabled As Boolean = True
    <DebuggerStepThrough()> _
    Protected Overrides Sub WndProc(ByRef m As Message)
        Const WM_PAINT As Int32 = &HF
        If m.Msg = WM_PAINT AndAlso Not PaintEventEnabled Then
            m.Result = IntPtr.Zero
        ElseIf m.Msg = WM_MOUSEWHEEL Then
            If (m.WParam.ToInt32 And MK_CONTROL) = MK_CONTROL Then
                m.WParam = IntPtr.Zero
                m.Result = IntPtr.Zero
            Else
                MyBase.WndProc(m)
            End If
        Else
            MyBase.WndProc(m)
        End If
    End Sub


    ' ===============================================================================================
    '  FIND LABEL
    ' ===============================================================================================
    Friend Function FindLabel(ByVal lines() As String, ByVal TextToFind As String) As Int32
        If TextToFind Is Nothing Then Return -1
        ' ----------------------------------------------------------------------- prepare text to find
        Dim TextToFind1 As String = " " + TextToFind.ToLower
        Dim TextToFind2 As String = " """ + TextToFind.ToLower + """"
        Dim TextToFind3 As String = " " & ExtractFirstWord(TextToFind).ToLower
        Dim TextToFind4 As String = " " & EvalString_AutoQuoting(TextToFind).ToLower
        Dim line As String
        For i As Int32 = 0 To lines.Length - 1
            line = lines(i).Trim.ToLower
            If line.StartsWith("label") Then
                RemoveComments(line)
                If line.EndsWith(TextToFind1) Or _
                   line.EndsWith(TextToFind2) Or _
                   line.EndsWith(TextToFind3) Or _
                   line.EndsWith(TextToFind4) Then
                    Return i
                End If
            End If
        Next
        ' ------------------------------------------------------------------------ not found
        Return -1
    End Function


    ' ===============================================================================================================
    '   RTB - SAVE and RESTORE selection and scroll ( scroll aligned with the first text line )
    ' ===============================================================================================================
    Friend Structure ScrollData
        Dim OldSelectionStart As Int32
        Dim OldSelectionLength As Int32
        Dim OldSelectionFirstChar As Int32
        Dim OldSelectionLastChar As Int32
    End Structure

    Friend Sub SaveSelectionAndScroll(ByRef sd As ScrollData)
        sd.OldSelectionStart = Me.SelectionStart
        sd.OldSelectionLength = Me.SelectionLength
        sd.OldSelectionFirstChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        sd.OldSelectionLastChar = Me.GetCharIndexFromPosition(New Point(1, Me.ClientSize.Height))
    End Sub

    Friend Sub RestoreSelectionAndScroll(ByRef sd As ScrollData)
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
        ' ---------------- CHANGED in VERSION 4.5 23/09/2019 ------------------- 
        ' The first version restores the start position to zero in some cases
        ' For example adding a "p" at the end of : Label pppp  
        ' (if pppp is the last file charachter)
        ' -------------------------------------------------------------- original
        'If sd.OldSelectionStart >= sd.OldSelectionFirstChar And _
        '   sd.OldSelectionStart <= sd.OldSelectionLastChar + 1 Then
        '    Me.Select(sd.OldSelectionStart, sd.OldSelectionLength)
        'End If
        ' -------------------------------------------------------------- changed
        If sd.OldSelectionStart >= sd.OldSelectionFirstChar Then
            Me.Select(sd.OldSelectionStart, sd.OldSelectionLength)
        End If
    End Sub

    ' ===============================================================================================
    '  LINES COLORS
    ' ===============================================================================================
    Friend ColorAllLinesMilliseconds As Int64

    Friend Sub ColorAllLines()
        ' ----------------------------------------------------------
        Dim sw As Stopwatch = New Stopwatch
        sw.Start()
        ' ----------------------------------------------------------
        RTBLines = Me.Lines
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTBLines.Length - 1
            ' ------------------------------------------------------ AutomationLanguageColors
            ColorSingleLine_AndReplaceKeywords(i)
            ' ------------------------------------------------------ Progress indicator
            If i Mod 100 = 0 Then
                ' --------------------------------------------------
                Form1.StatusLabel1.Text = " Line: " + i.ToString
                Form1.StatusLabel1.Refresh()
                ' -------------------------------------------------- 
                If sw.ElapsedMilliseconds > 2000 Then
                    Form1.Opacity = 1
                    Form1.Show()
                End If
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
        ' ---------------------------------------------------------- load time
        ColorAllLinesMilliseconds = sw.ElapsedMilliseconds
    End Sub

    Private Sub ColorAllLines_StartingWithVariable()
        ' ----------------------------------------------------------
        RTBLines = Me.Lines
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Int32 = 0 To RTBLines.Length - 1
            If RTBLines(i).ToLower.StartsWith("variable") Then
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
    End Sub

    Friend Sub ColorAllLines_StartingWith(ByVal str As StringCollection)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTBLines.Length - 1
            Dim ok As Boolean = False
            For j As Int32 = 0 To str.Count - 1
                If RTBLines(i).TrimStart.ToLower.StartsWith(str.Item(j)) Then
                    ok = True
                End If
            Next
            If ok Then
                SelectLine(i)
                Me.SelectionColor = Color_Text
                Me.SelectionBackColor = Color_ActiveBack
                ' ------------------------------------------------------ AutomationLanguageColors
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
    End Sub

    Friend Sub ColorAllLines_Containing(ByVal str As StringCollection)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTBLines.Length - 1
            Dim ok As Boolean = False
            For j As Int32 = 0 To str.Count - 1
                If RTBLines(i).TrimStart.ToLower.Contains(str.Item(j)) Then
                    ok = True
                End If
            Next
            If ok Then
                SelectLine(i)
                Me.SelectionColor = Color_Text
                Me.SelectionBackColor = Color_ActiveBack
                ' ------------------------------------------------------ AutomationLanguageColors
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
    End Sub

    Friend Sub ColorAllLines_StartingWith_And_Containing(ByVal strStarting As StringCollection, _
                                                         ByVal strContaining As StringCollection)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = False
        Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Dim l As String
        For i As Integer = 0 To RTBLines.Length - 1
            Dim ok As Boolean = False
            For j As Int32 = 0 To strStarting.Count - 1
                l = RTBLines(i).TrimStart.ToLower
                If l.StartsWith(strStarting.Item(j)) Then
                    For k As Int32 = 0 To strContaining.Count - 1
                        If l.Contains(strContaining.Item(k)) Then
                            ok = True
                            Exit For
                        End If
                    Next
                End If
                If ok Then Exit For
            Next
            If ok Then
                SelectLine(i)
                Me.SelectionColor = Color_Text
                Me.SelectionBackColor = Color_ActiveBack
                ' ------------------------------------------------------ AutomationLanguageColors
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Enabled = True
        PaintEventEnabled = True
        Me.Focus()
    End Sub

    Private Sub ColorLines(ByVal startline As Int32, ByVal endline As Int32)
        RTBLines = Me.Lines
        For i As Int32 = startline To endline
            ColorSingleLine(i)
        Next
    End Sub

    Private Sub ColorSingleLine(ByVal i As Int32)
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        ' ---------------------------------------------------------- save position 
        Dim OldCursorPosition As Integer = SelectionStart
        ' ---------------------------------------------------------- set the whole line to default colors
        SelectLine(i)
        Me.SelectionColor = Color_Text
        Me.SelectionBackColor = Color_ActiveBack
        ' ---------------------------------------------------------- AutomationLanguageColors
        ColorSingleLine_AndReplaceKeywords(i)
        ' ---------------------------------------------------------- restore
        SelectionStart = OldCursorPosition
        SelectionLength = 0
        ' ---------------------------------------------------------- 
        EventsAreEnabled = old
    End Sub

    Private Sub ColorSelectedLine()
        RTBLines = Me.Lines
        ColorSingleLine(GetLineFromCharIndex(SelectionStart))
    End Sub


    ' ===============================================================================================
    '   ON TEXT CHANGED
    ' ===============================================================================================
    Private LastEditedLine As Int32 = -1
    Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
        If Not PaintEventEnabled Then Return
        ' ------------------------------------------------------- 
        RTBLines = Me.Lines
        ' ------------------------------------------------------- 
        Dim line As Int32 = GetLineFromCharIndex(SelectionStart)
        ' ------------------------------------------------------- stop Paint
        PaintEventEnabled = False
        ' No "Enabled = false" here
        ' otherwise the key autorepeat stops and the scroll bar flashes
        ' ------------------------------------------------------- process the line
        ColorSingleLine(line)
        ' ------------------------------------------------------- eventually process also the precedent line
        If line <> LastEditedLine Then
            ColorSingleLine(line - 1)
        End If
        ' ------------------------------------------------------- restart Paint
        PaintEventEnabled = True
        ' ------------------------------------------------------- 
        TextModified = True
        TextMustBeSaved = True
        TextMustBeIndented = True
        ' ------------------------------------------------------- LastEditedLine <= 0 for undo deleting all the text
        If line <> LastEditedLine Or LastEditedLine <= 0 Then
            UndoAndHistory_AddEntry()
            LastEditedLine = line
        End If
    End Sub

    ' ===============================================================================================
    '  TEST KEYWORDS AND EVENTUALLY RECOLOR LINES 
    ' ===============================================================================================
    Private TestLines_StartingWords As StringCollection = New StringCollection
    Private TestLines_ContainingWords As StringCollection = New StringCollection

    Private Sub TestKeywords_PrepareKeywordsToRecolor()
        ' ---------------------------------------------------------------
        '  This function is called from KEYDOWN AND KEYUP
        ' ---------------------------------------------------------------
        Dim startline As Int32 = GetFirstSelectedLine()
        If startline < 0 Then Return
        If startline >= RTBLines.Length Then Return
        Dim endline As Int32 = GetLastSelectedLine()
        For i As Int32 = startline To endline
            TestKeywords_PrepareKeywordsToRecolor(i)
        Next
    End Sub

    Private Sub TestKeywords_PrepareKeywordsToRecolor(ByVal nline As Int32)
        Dim s As String = RTBLines(nline).Trim.ToLower
        If s.StartsWith("next") Then
            TestLines_StartingWords.Add("for")
            TestLines_StartingWords.Add("exit")
        End If
        If s.StartsWith("end") Or s.StartsWith("return") Then
            TestLines_StartingWords.Add("label")
        End If
        If s.StartsWith("label") Then
            AddString_NewStringOnly(TestLines_StartingWords, "button")
            AddString_NewStringOnly(TestLines_StartingWords, "key")
            AddString_NewStringOnly(TestLines_StartingWords, "gosub")
            AddString_NewStringOnly(TestLines_StartingWords, "goto")
            If s.Length > 5 Then
                Dim word As String = s.Substring(6)
                RemoveComments(word)
                AddString_NewStringOnly(TestLines_ContainingWords, word.Trim.ToLower.Replace("""", ""))
            End If
        End If
        If s.StartsWith("else") Then
            AddString_NewStringOnly(TestLines_StartingWords, "if")
        End If
        If s.StartsWith("endif") Then
            AddString_NewStringOnly(TestLines_StartingWords, "if")
            AddString_NewStringOnly(TestLines_StartingWords, "else")
        End If
        If s.StartsWith("endselect") Then
            AddString_NewStringOnly(TestLines_StartingWords, "select")
            AddString_NewStringOnly(TestLines_StartingWords, "case")
        End If
        If s.StartsWith("variable") Then
            Dim sa() As String = ReplaceMultipleSpacesAndTabAndTrim(s).Split(" "c)
            If sa.Length >= 3 Then
                If sa(2).Length >= UserVariablesMinLength Then
                    AddString_NewStringOnly(TestLines_ContainingWords, sa(2))
                End If
            End If
        End If
        ' --------------------------------------------- test for unesed variables (could be slow)
        If TheLineContainsSomeUserVariable(s) Then
            If Not s.StartsWith("variable") Then
                AddString_NewStringOnly(TestLines_StartingWords, "variable")
            End If
        End If
    End Sub

    Private Sub AddString_NewStringOnly(ByRef sc As StringCollection, ByVal s As String)
        If Not sc.Contains(s) Then
            sc.Add(s)
        End If
    End Sub

    Private Sub TestKeywords_RecolorLines()
        ' ---------------------------------------------------------------
        '  This function is called from KEYUP
        ' ---------------------------------------------------------------
        RTBLines = Me.Lines()  ' <--- 900 uS
        ' ----------------------------------------------------------------
        Cursor.Current = Cursors.WaitCursor
        If TestLines_StartingWords.Count > 0 And TestLines_ContainingWords.Count > 0 Then
            ColorAllLines_StartingWith_And_Containing(TestLines_StartingWords, TestLines_ContainingWords)
            TestLines_StartingWords.Clear()
            TestLines_ContainingWords.Clear()
        ElseIf TestLines_StartingWords.Count > 0 Then
            ColorAllLines_StartingWith(TestLines_StartingWords)
            TestLines_StartingWords.Clear()
        ElseIf TestLines_ContainingWords.Count > 0 Then
            ColorAllLines_Containing(TestLines_ContainingWords)
            TestLines_ContainingWords.Clear()
        End If
        Cursor.Current = Cursors.Default
    End Sub

    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  COLOR SINGLE LINE AND REPLACE KEYWORDS
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################
    Private LineWithErrors As Boolean

    Private Sub ColorSingleLine_AndReplaceKeywords(ByVal nline As Int32)
        If nline < 0 Or nline > RTBLines.Length - 1 Then Return
        ' --------------------------------------------------------------- 
        Dim line As String = RTBLines(nline)
        ' --------------------------------------------------------------- 
        If line <> "" Then
            ' ----------------------------------------------------------- color active back
            SelectLine(nline)
            Me.SelectionColor = Color_Text
            Me.SelectionBackColor = Color_ActiveBack
            ' ----------------------------------------------------------- comment position
            Dim CommentPosition As Int32 = RemoveComments(line)
            If CommentPosition = 0 Then
                MarkCommentLine(nline, 0)
                Return
            End If
            '
            LineWithErrors = False
            ' ----------------------------------------------------------- tokenize to array of keywords
            Dim sa() As String = Tokenize(line.Trim.ToLowerInvariant)
            ' ----------------------------------------------------------- 
            Select Case sa.Length
                Case 0
                Case 1
                    Select Case sa(0)
                        Case "beep", "controls", "next", "case", "caseelse", "endif", "endselect", "return", "select", "stop", "end"
                        Case "exit"
                            If Not TestNext(nline) Then MarkNotFoundLine(nline)
                        Case "else"
                            If Not TestEndif(nline) Then MarkNotFoundLine(nline)
                        Case Else : MarkErrorLine(nline)
                    End Select
                Case 2
                    Select Case sa(0)
                        Case "case"
                            If Not TestEndselect(nline) Then MarkNotFoundLine(nline)
                            If Not IsValidBooleanStringOrDouble(sa(1)) Then MarkErrorLine(nline)
                        Case "com"
                            If sa(1) <> "close" AndAlso _
                               sa(1) <> "discardinbuffer" AndAlso _
                               sa(1) <> "discardoutbuffer" AndAlso _
                               sa(1) <> "enabledtr" AndAlso _
                               sa(1) <> "enablerts" AndAlso _
                               sa(1) <> "disabledtr" AndAlso _
                               sa(1) <> "disablerts" Then
                                MarkErrorLine(nline)
                            End If
                        Case "controls"
                            If sa(1) <> "opentextbox" AndAlso _
                               sa(1) <> "closetextbox" AndAlso _
                               sa(1) <> "cleartextbox" Then
                                MarkErrorLine(nline)
                            End If
                        Case "if"
                            If Not TestElseOrEndif(nline) Then MarkNotFoundLine(nline)
                            If Not IsValidBoolean(sa(1)) Then MarkErrorLine(nline)
                        Case "wait"
                            If Not IsValidBoolean(sa(1)) Then MarkErrorLine(nline)
                        Case "label"
                            If FindLabel(RTBLines, sa(1)) <> nline Then
                                MarkErrorLine(nline)
                            Else
                                If sa(1) = "eventstop" Then
                                    If Not TestTheFinalEndForTheEventStopLabel(nline) Then
                                        MarkErrorLine(nline)
                                    End If
                                ElseIf sa(1) = "eventtimer" Then
                                    Form1.EventTimer_Init()
                                    If Form1.EventTimer_StartLine < 0 Then
                                        MarkErrorLine(nline)
                                    End If
                                End If
                            End If
                        Case "gosub", "goto"
                            If FindLabel(RTBLines, sa(1)) < 0 Then
                                If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                    MarkWarningLine(nline)
                                Else
                                    MarkNotFoundLine(nline)
                                End If
                            End If
                        Case "load"
                            If sa(1) <> "slots" AndAlso _
                               sa(1) <> "vars" AndAlso _
                               sa(1) <> "stop" AndAlso _
                               sa(1) <> "pause" AndAlso _
                               sa(1) <> "play" AndAlso _
                               sa(1) <> "hide" AndAlso _
                               sa(1) <> "closeapps" Then
                                If sa(1).ToLower.Trim.StartsWith("slots") Then
                                    Dim n() As Double = ExtractNumericParams(sa(1), 3)
                                    If n.Length <> 2 Then MarkErrorLine(nline)
                                Else
                                    '
                                    If sa(1).StartsWith("position") Then
                                        Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(1)).Split(" "c)
                                        If sa2.Length <> 2 Then
                                            MarkErrorLine(nline)
                                        End If
                                    End If
                                    '
                                    Dim str As String = EvalString_AutoQuoting(sa(1))
                                    Select Case Strings.Right(str, 4).ToLower
                                        Case ".txt"
                                            '
                                            ' try to load any string (if exists) 
                                            ' 
                                            ' --------------------------------------------- test load file
                                            'Dim sa2() As String = sa(1).Split(" "c)
                                            'If sa2.Length > 1 Then
                                            '    If Not IsStringVariable(sa2(0)) Then MarkErrorLine(nline)
                                            'End If

                                            If Not IsValidString_AutoQuoting(str) Then MarkErrorLine(nline)
                                    End Select
                                End If
                            End If
                        Case "presskeys"
                            If Not IsValidString_AutoQuoting(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "sendkeys"
                            If Not IsValidString_AutoQuoting(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "print"
                            If Not IsValidBooleanStringOrDouble(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "save"
                            If sa(1) <> "slots" AndAlso _
                               sa(1) <> "vars" Then
                                MarkErrorLine(nline)
                            End If
                        Case "select"
                            If Not TestCase(nline) Then MarkNotFoundLine(nline)
                            If Not TestEndselect(nline) Then MarkNotFoundLine(nline)
                            If Not IsValidBooleanStringOrDouble(sa(1)) Then MarkErrorLine(nline)
                        Case "speed"
                            If Not IsValidDouble(sa(1)) Then MarkErrorLine(nline)
                        Case "window"
                            If sa(1) <> "minsize" AndAlso _
                               sa(1) <> "sizable" AndAlso _
                               sa(1) <> "maximized" AndAlso _
                               sa(1) <> "fullscreen" Then
                                MarkErrorLine(nline)
                            End If
                        Case Else : MarkErrorLine(nline)
                    End Select
                Case 3
                    Select Case sa(0)
                        Case "button"
                            Select Case sa(2)
                                Case "enabled", "disabled"
                                Case "color" : MarkWarningLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                            If Not IsValidButtonNumber(sa(1)) And Not IsValidDouble(sa(1)) Then MarkErrorLine(nline)
                        Case "com"
                            Select Case sa(1)
                                Case "writestring", "writeline"
                                    If Not IsValidString_AutoQuoting(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "save"
                            Select Case sa(1)
                                Case "deletefile"
                                    If Not IsValidString_AutoQuoting(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "variable"
                            Select Case sa(1)
                                Case "numeric"
                                    If Not IsNumericUserVariable(sa(2), nline) Then MarkErrorLine(nline) : Exit Select
                                    If UnusedVariable(sa(2), nline) Then MarkWarningLine(nline)
                                Case "string"
                                    If Not IsStringUserVariable(sa(2), nline) Then MarkErrorLine(nline) : Exit Select
                                    If UnusedVariable(sa(2), nline) Then MarkWarningLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "wait"
                            Select Case sa(1)
                                Case "button"
                                    '
                                Case "seconds"
                                    If Not IsValidDouble(sa(2)) Then MarkErrorLine(nline)
                                Case "application"
                                    If Not IsValidString(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case Else
                            If sa(1) = "=" Then
                                If IsNumericVariable(sa(0)) Then
                                    If IsValidDouble(sa(2)) OrElse IsValidBoolean(sa(2)) Then
                                    Else
                                        MarkErrorLine(nline)
                                    End If
                                ElseIf IsStringVariable(sa(0)) Then
                                    If IsValidString(sa(2)) Then
                                    Else
                                        MarkErrorLine(nline)
                                    End If
                                Else
                                    MarkErrorLine(nline)
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                    End Select
                Case 4
                    Select Case sa(0)
                        Case "button"
                            If Not IsValidButtonNumber(sa(1)) And Not IsValidDouble(sa(1)) Then MarkErrorLine(nline)
                            Select Case sa(2)
                                Case "slot"
                                Case "text"
                                    If FindLabel(RTBLines, sa(3)) < 0 Then
                                        If sa(3).Contains(" ") Or IsSomeVariable(sa(3)) Then
                                            MarkWarningLine(nline)
                                        Else
                                            If sa(3).Replace("""", "").Length > 0 Then
                                                MarkWarningLine(nline)
                                            End If
                                        End If
                                    End If
                                Case "color"
                                    Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(3)).Split(" "c)
                                    If sa(3).StartsWith("flashing") Then
                                        If sa2.Length < 3 Then
                                            MarkWarningLine(nline)
                                        ElseIf sa2.Length = 3 Then
                                            If ColorFromColorName(sa2(1)) = Color.Empty Or _
                                               ColorFromColorName(sa2(2)) = Color.Empty Then
                                                MarkWarningLine(nline)
                                            End If
                                        Else
                                            MarkErrorLine(nline)
                                        End If
                                    Else
                                        If sa2.Length < 2 Then
                                            If ColorFromColorName(sa(3)) = Color.Empty Then
                                                MarkWarningLine(nline)
                                            End If
                                        Else
                                            MarkErrorLine(nline)
                                        End If
                                    End If
                                Case Else
                                        MarkErrorLine(nline)
                            End Select
                        Case "com"
                            If sa(1) = "open" Then
                                If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                                Else
                                    MarkErrorLine(nline)
                                End If
                            End If
                        Case "key"
                            If IsValidKey(sa(1)) AndAlso (sa(2) = "goto" Or sa(2) = "gosub") Then
                                If FindLabel(RTBLines, sa(3)) < 0 Then
                                    If sa(3).Contains(" ") Or IsSomeVariable(sa(3)) Then
                                        MarkWarningLine(nline)
                                    Else
                                        MarkNotFoundLine(nline)
                                    End If
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "save"
                            Select Case sa(1)
                                Case "stringtofile"
                                    If Not IsValidString_AutoQuoting(sa(2)) And Not IsValidString_AutoQuoting(sa(3)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "slot"
                            If IsValidDouble(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) Then
                            Else
                                Select Case sa(3)
                                    Case "reset", "sleep"
                                    Case Else : MarkErrorLine(nline)
                                End Select
                            End If
                        Case Else
                            Select Case sa(2)
                                Case "input"
                                Case Else
                                    MarkErrorLine(nline)
                            End Select
                    End Select
                Case 6
                    If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                       IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) Then
                        If Not TestForNext(nline) Then
                            MarkErrorLine(nline)
                        End If
                    Else
                        MarkErrorLine(nline)
                    End If
                Case 8
                    If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                       IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) And _
                       sa(6) = "step" And IsValidDouble(sa(7)) Then
                        If Not TestForNext(nline) Then
                            MarkErrorLine(nline)
                        End If
                    Else
                        MarkErrorLine(nline)
                    End If
                Case Else
                    MarkErrorLine(nline)
            End Select
            ' ----------------------------------------------------------- color the comment section
            If CommentPosition > -1 Then
                MarkCommentLine(nline, CommentPosition)
            End If
            ' ----------------------------------------------------------- replace keywords
            If Not LineWithErrors Then
                ReplaceKeywords(nline, RTBLines(nline), CommentPosition)
            End If
        End If
    End Sub

    Private Sub MarkErrorLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_ErrorBack
        LineWithErrors = True
    End Sub
    Private Sub MarkNotFoundLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_NotFoundBack
    End Sub
    Private Sub MarkWarningLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_WarningBack
    End Sub
    Private Sub MarkCommentLine(ByVal i As Int32, Optional ByVal FirstCommentedChar As Int32 = -1)
        SelectLine(i)
        If FirstCommentedChar > 0 Then
            SelectionLength -= FirstCommentedChar ' lenght must be changed before start
            SelectionStart += FirstCommentedChar  ' otherwise at the end of the file the lenght is shortened 
        End If
        Me.SelectionColor = Color_CommentText
        Me.SelectionBackColor = Color_CommentBack
        ' ------------------------------------------------------------------
        If RTBLines(i).EndsWith("'Breakpoint") Then
            SelectLine(i)
            SelectionLength = 11
            SelectionStart += RTBLines(i).Length - 11
            Me.SelectionColor = Color_BreakpointText
            Me.SelectionBackColor = Color_BreakpointBack
        End If
    End Sub

    ' ===============================================================================================
    '  MIXED FUNCTIONS
    ' ===============================================================================================
    Private Function IsValidButtonNumber(ByVal str As String) As Boolean
        Dim n As Int32 = CInt(Val(str))
        If n >= 1 And n <= 32 Then Return True
        Return False
    End Function

    Friend Function IsValidKey(ByVal KeyString As String) As Boolean
        If KeyString = Nothing Then Return False
        Select Case KeyString
            Case "left", "right", "up", "down", "pageup", "pagedown", "space", "esc", "home", "end"
                Return True
            Case Else
                If KeyString.Length = 1 Then
                    If "0123456789abcdefghijklmnopqrstuvwxyz".Contains(KeyString) Then Return True
                ElseIf KeyString.Length = 2 Then
                    If "f1f2f3f4f5f6f7f8f9".Contains(KeyString) Then Return True
                ElseIf KeyString.Length = 3 Then
                    If "f10f11f12".Contains(KeyString) Then Return True
                End If
        End Select
        Return False
    End Function

    Friend Function ExtractFirstWord(ByVal s As String) As String
        If s = Nothing Then Return ""
        s = s.Replace(vbTab, "").Trim
        Dim i As Int32 = s.IndexOf(" ")
        If i < 0 Then Return s
        Return s.Substring(0, i)
    End Function

    Friend Function RemoveComments(ByRef s As String) As Int32
        Dim i As Int32 = s.IndexOf("'")
        If i = 0 Then s = ""
        If i > 0 Then
            i = s.Length
            Dim InsideString As Boolean = False
            For j As Int32 = 0 To s.Length - 1
                Select Case s(j)
                    Case """"c : InsideString = Not InsideString
                    Case "'"c
                        If Not InsideString Then
                            i = j
                            s = s.Remove(i).TrimEnd
                            Exit For
                        End If
                End Select
            Next
        End If
        Return i
    End Function

    Friend Sub SelectLine(ByVal line As Int32)
        If line >= 0 AndAlso line < RTBLines.Length AndAlso RTBLines.Length > 0 Then
            Dim lineStart As Int32 = GetFirstCharIndexFromLine(line)
            Dim lineLength As Int32 = RTBLines(line).Length
            Me.Select(lineStart, lineLength)
        Else
            Me.Select(0, 0)
        End If
    End Sub
    Friend Sub SelectLines(ByVal FirstLine As Int32, ByVal LastLine As Int32)
        Dim FirstChar As Int32 = Me.GetFirstCharIndexFromLine(FirstLine)
        Dim LastChar As Int32 = Me.GetFirstCharIndexFromLine(LastLine) + RTBLines(LastLine).Length - 1
        Me.Select(FirstChar, LastChar - FirstChar)
    End Sub
    Friend Function GetFirstSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart)
    End Function
    Private Function GetLastSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart + Me.SelectionLength)
    End Function
    Friend Function GetFirstVisibleLine() As Int32
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(New Point(1, 1)))
    End Function
    Friend Function GetLastVisibleLine() As Int32
        Dim p As Point = New Point(1, Me.ClientSize.Height)
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(p))
    End Function
    Friend Sub ShowLinesCount()
        RTBLines = Me.Lines
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        Dim nCol As Int32 = GetColumnFromCharIndex(SelectionStart)
        Form1.Label_Lines.Text = "Lines: " + RTBLines.Length.ToString + _
                                 "  Line: " + (nline + 1).ToString + _
                                 "  Col: " + nCol.ToString
    End Sub

    Friend Function GetLastCharIndexFromLine(ByVal nline As Int32) As Int32
        Return Me.GetFirstCharIndexFromLine(nline) + RTBLines(nline).Length
    End Function

    Private Function GetLineLengthAsEquivalentSpaces(ByVal nline As Int32) As Int32
        'RTBLines = Me.Lines
        If RTBLines.Length = 0 Then Return 0
        Return GetColumnFromCharIndex(GetFirstCharIndexFromLine(nline) + RTBLines(nline).Length)
    End Function

    ' -------------------------------- 15 uS (independent from n-lines) -------
    Private Function GetColumnFromCharIndex(ByVal charindex As Int32) As Int32
        Dim n As Int32 = 0
        Dim nline As Int32 = Me.GetLineFromCharIndex(charindex)
        Dim nchars As Int32 = charindex - GetFirstCharIndexFromLine(nline)
        For i As Int32 = 0 To nchars - 1
            If RTBLines(nline)(i) = CChar(vbTab) Then
                n += 4 - (n Mod 4)
            Else
                n += 1
            End If
        Next
        Return n
    End Function

    Friend Sub ScrollToCenterSelectedLine(ByVal line As Int32)
        Dim VisibleLinesCount As Int32 = GetLastVisibleLine() - GetFirstVisibleLine()
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        PaintEventEnabled = False
        Enabled = False
        Dim delta As Int32 = VisibleLinesCount \ 2 - 1
        ScrollToLine(line + delta)
        ScrollToLine(line - delta)
        Enabled = True
        PaintEventEnabled = True
        EventsAreEnabled = old
    End Sub

    Friend Sub ScrollToLine(ByVal nline As Int32)
        If RTBLines.Length = 0 Then Return
        If nline < 0 Then nline = 0
        If nline > RTBLines.Length - 1 Then nline = RTBLines.Length - 1
        Me.Select(GetFirstCharIndexFromLine(nline), 1)
        Me.ScrollToCaret()
    End Sub

    Private Function TestEndif(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("endif") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("if") Then
                counter += 1
            End If
        Next
        Return False
    End Function
    Private Function TestElseOrEndif(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("else") Then
                If counter <= 0 Then
                    Return True
                End If
            ElseIf line.StartsWith("endif") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("if") Then
                counter += 1
            End If
        Next
        Return False
    End Function
    Private Function TestCase(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("case") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("select") Then
                counter += 1
            ElseIf line.StartsWith("endselect") Then
                counter -= 1
            End If
        Next
        Return False
    End Function
    Private Function TestEndselect(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("endselect") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("select") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    Private Function TestForNext(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("next") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("for") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    Private Function TestTheFinalEndForTheEventStopLabel(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("end") And Not _
               line.StartsWith("endif") And Not _
               line.StartsWith("endselect") Then
                Return True
            ElseIf line.StartsWith("label") Or _
                    line.StartsWith("return") Then
                Return False
            End If
        Next
        Return False
    End Function

    Private Function TestNext(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTBLines.Length - 1
            line = RTBLines(i).Trim.ToLower
            If line.StartsWith("next") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("for") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    Friend Sub AddFinalSpace_ToRow()
        Dim s As String = RTBLines(GetFirstSelectedLine())
        If Not s.EndsWith(" ") And Not s.EndsWith(vbTab) Then
            SelectionLength = 0
            SelectedText = " "
        End If
    End Sub


    ' ===============================================================================================
    '  REPLACE KEYWORDS
    ' ===============================================================================================

    Private Keywords() As String = {"If", "Int", "And", "Or", "Xor", "Not", _
                                    "CRLF", "CR", "LF", "NBSP", "TAB", "CLS", "Math_PI", "Math_E", _
                                    "Chars", "Chr", "Asc", "Bin", "FromBin", "Hex", "FromHex", _
                                    "Sin", "Cos", "Asin", "Acos", "Tan", "Atan", "Atan2", "Pow", "Sqrt", _
                                    "Round", "Mid", "Len", "Abs", "Sign", "Rnd", "Print", "EventStop", "EventTimer", _
                                    "Disabled", "Enabled", "Color", "Input", "True", "False", "Reset", "Key", _
                                    "Seconds", "Sleep", "Slot", "End", "Exit", "Text", "Hide", "ElapsedTime", "Wait", _
                                    "Trim", "Str", "Format", "UCase", "LCase", "WCase", "Date", "Now", "Today", _
                                    "PadLeft", "PadRight", "COM_Status", "COM_Received", "COM_PortNames", _
                                    "GetSeparatedStringCount", "GetSeparatedString", "Application", _
                                    "Replace", "RemoveComments", "OpenTextBox", "CloseTextBox", "ClearTextBox", _
                                    "Select", "Case", "CaseElse", "MouseX", "MouseY", "PressKeys", "SendKeys"}

    Friend Sub ReplaceKeywords(ByVal lineNumber As Int32, _
                               ByRef line As String, _
                               ByVal commentPosition As Int32)
        Dim LineStart As Int32
        If lineNumber >= 0 Then LineStart = GetFirstCharIndexFromLine(lineNumber)
        ' --------------------------------------------------------------- SPECIAL LINES
        Dim line2 As String = line.Trim.ToLower
        '
        If line2.StartsWith("beep") Then
            ReplaceKeywords(New String() {"Beep"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("com ") Then
            ReplaceKeywords(New String() {"COM", "CRLF", "Open", "Close", "WriteString", "WriteLine", _
                                          "DiscardOutBuffer", "DiscardInBuffer", _
                                          "EnableDTR", "EnableRTS", "DisableDTR", "DisableRTS"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("controls ") Then
            ReplaceKeywords(New String() {"Controls", "OpenTextBox", "CloseTextBox", "ClearTextBox"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("button ") Then
            If line2.Contains("text") Then
                ReplaceKeywords(New String() {"Button", "Text"}, lineNumber, LineStart, line, commentPosition)
            ElseIf line2.Contains("color") Then
                ReplaceKeywords(New String() {"Button", "Color"}, lineNumber, LineStart, line, commentPosition)
            Else
                ReplaceKeywords(New String() {"Button", "Slot", "Enabled", "Disabled"}, lineNumber, LineStart, line, commentPosition)
            End If
            If line2.Contains("color") Then
                ReplaceKeywords(GetColorNames.Split(" "c), lineNumber, LineStart, line, commentPosition, True)
                ReplaceKeywords(New String() {"Flashing", "Flashing1", "Flashing2", "Flashing3", "Flashing4", _
                                              "Flashing5", "Flashing6", "Flashing7", "Flashing8", "Flashing9"}, lineNumber, LineStart, line, commentPosition, True)
            End If

        ElseIf line2.StartsWith("for ") Then
            ReplaceKeywords(New String() {"For", "To", "Step"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("next") Then
            ReplaceKeywords(New String() {"Next"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("gosub ") Then
            ReplaceKeywords(New String() {"Gosub"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("goto ") Then
            ReplaceKeywords(New String() {"Goto"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("return") Then
            ReplaceKeywords(New String() {"Return"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("else") Then
            ReplaceKeywords(New String() {"Else"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("endif") Then
            ReplaceKeywords(New String() {"EndIf"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("key ") Then
            ReplaceKeywords(New String() {"Key", "Goto", "Gosub", "Left", "Right", "Up", "Down", "Pageup", "Pagedown", "Space", "Esc", "Home", "End"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("label ") Then
            ReplaceKeywords(New String() {"Label", "EventStop", "EventTimer"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("load ") Then
            ReplaceKeywords(New String() {"Load", "Play", "Pause", "Position", "Stop", "Hide", "Slots", "Vars", "CloseApps"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("presskeys ") Then
            ReplaceKeywords(New String() {"PressKeys"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetPressKeysNames_NoLetters().Split(" "c), lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("sendkeys ") Then
            ReplaceKeywords(New String() {"SendKeys"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetSendKeysNames_NoLetters().Split(" "c), lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("save ") Then
            ReplaceKeywords(New String() {"Save", "Slots", "Vars", "StringToFile", "DeleteFile"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("speed ") Then
            ReplaceKeywords(New String() {"Speed"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("stop") Then
            ReplaceKeywords(New String() {"Stop"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("caseelse") Then ' <-- must be before the "case"
            ReplaceKeywords(New String() {"CaseElse"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("endselect") Then
            ReplaceKeywords(New String() {"EndSelect"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("variable ") Then
            ReplaceKeywords(New String() {"Variable", "Numeric", "String"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("window ") Then
            ReplaceKeywords(New String() {"Window", "MinSize", "Sizable", "Maximized", "FullScreen"}, lineNumber, LineStart, line, commentPosition)

        Else
            ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)                  ' REPLACE ALL KEYWORDS
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        End If
    End Sub

    Private Sub ReplaceKeywords(ByVal Keywords As String(), _
                                ByVal lineNumber As Int32, _
                                ByVal lineStart As Int32, _
                                ByRef line As String, _
                                ByVal commentPosition As Int32, _
                                Optional ByVal UseTextColor As Boolean = False)

        If commentPosition < 0 Then commentPosition = 9999
        Dim position As Int32
        For Each kw As String In Keywords
            If kw = "" Then Continue For
            position = 0
            Do
                position = line.IndexOf(kw, position, StringComparison.InvariantCultureIgnoreCase)
                If position >= 0 Then
                    If IsCompleteWord(line, position, kw.Length) And position < commentPosition Then
                        If Not InsideQuotes(position, line) Then
                            If lineNumber >= 0 Then
                                Me.Select(lineStart + position, kw.Length)
                                Me.SelectionColor = If(UseTextColor, Color_Text, Color_KeywordText)
                                Me.Select(lineStart + position, kw.Length) ' <-- redo selection after SelectionColor 
                                If kw <> Me.SelectedText Then
                                    Me.SelectedText = kw
                                End If
                            Else
                                Mid(line, position + 1, kw.Length) = kw
                            End If
                        End If
                    End If
                    position += kw.Length
                End If
            Loop Until position < 0
        Next
    End Sub

    Private Function IsCompleteWord(ByVal l As String, ByVal position As Int32, ByVal length As Int32) As Boolean
        ' ------------------------------------------------------------ test left side
        If position > 1 AndAlso _
           IsLetterOrSpaceOrNumber(l(position - 1)) Then Return False
        ' ------------------------------------------------------------- test right side
        If position + length < l.Length AndAlso _
           IsLetterOrSpaceOrNumber(l(position + length)) Then Return False
        ' -------------------------------------------------------------
        Return True
    End Function

    Private Function IsLetterOrSpaceOrNumber(ByVal c As Char) As Boolean
        Return (c >= "a" And c <= "z") OrElse (c >= "A" And c <= "Z") OrElse (c >= "0" And c <= "9") OrElse c = "_"
    End Function

    Private Function InsideQuotes(ByVal index As Int32, ByVal text As String) As Boolean
        InsideQuotes = False
        If index >= text.Length Then Return False
        For Counter As Int32 = 0 To index - 1
            If text(Counter) = """"c Then
                InsideQuotes = Not InsideQuotes
            End If
        Next
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  AUTO INDENT
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################
    Friend Sub AutoIndent()
        ' -----------------------------------------------------------
        If Not PaintEventEnabled Then Exit Sub ' <-- important
        PaintEventEnabled = False ' <-- important
        ' No "Enabled = false" here
        ' otherwise the key autorepeat stops and the scroll bar flashes
        ' ----------------------------------------------------------- save position
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        Dim SelectionModified As Boolean = False
        ' ----------------------------------------------------------- save the cursor line index
        Dim CursorLine As Int32 = GetFirstSelectedLine()
        ' ----------------------------------------------------------- 
        Dim indent As Int32 = 0
        For i As Integer = 0 To RTBLines.Length - 1
            ' -------------------------------------------------------
            Dim l() As String
            l = RTBLines(i).Replace(vbTab, " ").Trim.ToLower.Split(" "c)
            If l(0).StartsWith("'") Then Continue For

            ' ------------------------------------------------------- 
            Select Case l(0)
                Case "endif", "case", "caseelse", "else", "next" : indent -= 1
                Case "endselect" : indent -= 2
                Case "label" : If LineIsEmptyOrCommentOrTerminator(i - 1) Then indent = 0
                Case "goto", "return", "stop", "end"
                    If LineIsEmptyOrCommentOrTerminator(i + 1) Then indent -= 1
            End Select
            ' ------------------------------------------------------- 
            Dim ntabs As Int32 = indent
            If ntabs < 0 Then ntabs = 0
            ' ------------------------------------------------------- 
            Dim lineText As String = RTBLines(i)
            Dim InitialTabs As Int32 = CountInitialTabs(lineText, 4)
            If ntabs <> InitialTabs Then
                Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                Me.Select(start, CountInitialChars(lineText))
                Me.SelectedText = StrDup(ntabs, vbTab)
                SelectionModified = True
                ' -----------------------------------------------------------------
                '  if indenting the cursor line then correct the restore selection
                ' -----------------------------------------------------------------
                If i = CursorLine - 1 Then
                    Dim delta As Int32 = ntabs - InitialTabs
                    sd.OldSelectionStart += delta
                End If
                If i = CursorLine Then
                    If LastKeyPressed = Keys.Enter Then
                        sd.OldSelectionStart += ntabs
                    End If
                End If
                ' ---------------------------------------------------
            End If
            ' ------------------------------------------------------- 
            Select Case l(0)
                Case "if", "label", "case", "caseelse", "else", "for" : indent += 1
                Case "select" : indent += 2
            End Select
            ' ------------------------------------------------------- 
            If indent < 0 Then indent = 0
        Next
        ' ----------------------------------------------------------- restore selection and scroll
        If SelectionModified Then
            RestoreSelectionAndScroll(sd)
            TextModified = True
            TextMustBeSaved = True
        End If
        ' ----------------------------------------------------------- 
        PaintEventEnabled = True ' <-- important
        'Me.Focus()
        TextMustBeIndented = False
    End Sub

    Private Function LineIsEmptyOrCommentOrInitiator(ByVal index As Int32) As Boolean
        If index < 0 Then Return True
        If index >= RTBLines.Length Then Return True
        Dim l As String = RTBLines(index).Trim.ToLower
        If l = "" OrElse l.StartsWith("'") _
            OrElse l.StartsWith("label") Then Return True
        Return False
    End Function

    Private Function LineIsEmptyOrCommentOrTerminator(ByVal index As Int32) As Boolean
        If index < 0 Then Return True
        If index >= RTBLines.Length Then Return True
        Dim l As String = RTBLines(index).Trim.ToLower
        If l = "" OrElse l.StartsWith("'") _
            OrElse l.StartsWith("goto") _
            OrElse l.StartsWith("return") _
            OrElse l.StartsWith("stop") _
            OrElse (l.StartsWith("end") AndAlso Not l.StartsWith("endif")) Then Return True
        Return False
    End Function



    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  SUGGESTIONS
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Friend SuggestionsButtonType As String = ""
    Private AllKeywords As String = "Beep  Button  Controls  For  Next  Exit  GoSub  GoTo  If  Else  EndIf  Key  Label  Load  PressKeys  SendKeys  Print  Return  Save  Select  Case  CaseElse  EndSelect  Slot  Speed  Stop  End  Variable  Wait  Window"
    Private LoadKeywords As String = "Slots Vars Stop Pause Play Position Hide CloseApps  or  FILENAME(.exe, .txt, .jpg, .bmp, .png, .gif, .avi, .wmv, .mpg, .mp4, .mov, .mp3, .wav)  or  HTTP  or  a generic string expression"
    Friend Sub ShowSuggestions()
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        ShowLinesCount()
        If RTBLines.Length = 0 Then
            Form1.StatusLabel1.BackColor = Color.FromArgb(220, 255, 200)
            Form1.SetStatusLabelFontProportional()
            Form1.StatusLabel1.Text = "Keywords:  " & AllKeywords
            Return
        End If
        ' ----------------------------------------------------------- 
        ShowSuggestions(RTBLines(nline), nline, GetCompleteWord(Me.Text, Me.SelectionStart))
    End Sub

    Friend Sub ShowSuggestions(ByVal _line As String, ByVal nline As Int32, ByVal SelectedWord As String)
        LastError = ""
        Dim line As String = _line.Replace(vbTab, " ").Trim
        RemoveComments(line)
        ' ----------------------------------------------------------- Tokenize to array of keywords 
        Dim sa() As String = Tokenize(line)
        ' ----------------------------------------------------------- sa2 contains the original case strings 
        Dim saCase() As String = CType(sa.Clone, String())
        ' ----------------------------------------------------------- set string aray sa case to lower
        For i As Int32 = 0 To sa.Length - 1
            If sa(i) IsNot Nothing Then sa(i) = sa(i).ToLower
        Next
        ' ----------------------------------------------------------- 
        SuggestionsButtonType = ""
        Dim s As String = ""
        Select Case sa.Length
            Case 0
                s = "Keywords: " + AllKeywords
            Case 1
                Select Case sa(0)
                    Case "beep" : s = "OK: A system Beep will be emitted here (repetition is limited to 10 per second)"
                    Case "button" : s = "NoKeywords: BUTTON-NUMBER"
                    Case "com" : s = "Keywords: Open / Close / WriteString / WriteLine / DiscardOutBuffer / DiscardInBuffer / EnableDTR / EnableRTS / DisableDTR / DisableRTS"
                    Case "controls" : s = "Keywords: OpenTextBox / CloseTextBox / ClearTextBox"
                    Case "for" : s = "NoKeywords: Please write a numeric variable (v1 to v9 or a user defined numeric variable)"
                    Case "next" : s = "OK: The execution returns to the preceding ""For"" (if exists)."
                    Case "exit"
                        If TestNext(nline) Then
                            s = "OK: The execution of the current FOR continues after the NEXT (if exists) ."
                        Else
                            s = "WARNING: This line has no effect without a corresponding Next"
                        End If
                    Case "gosub" : s = "NoKeywords: LABEL-NAME  (enclosing the text in double quotes is possible, but not necessary)"
                    Case "goto" : s = "NoKeywords: LABEL-NAME  (enclosing the text in double quotes is possible, but not necessary)"
                    Case "if" : s = "NoKeywords: EXPRESSION"
                    Case "else"
                        If TestEndif(nline) Then
                            s = "OK: Following instructions are executed when the preceding IF is FALSE"
                        Else
                            s = "WARNING: This line has no effect without a corresponding EndIf"
                        End If
                    Case "endif" : s = "OK: This is the end of the conditional execution."
                    Case "endselect" : s = "OK: This is the end of the Select/Case conditional execution."
                    Case "key" : s = "Keywords: 0..9 / a..z / F1..F12 / Left / Right / Up / Down / Pageup / Pagedown / Space / Esc / Home / End"
                    Case "label" : s = "NoKeywords: LABEL-NAME or ""EventStop"" or ""EventTimer""" + _
                                       "  ( the EventStop label will be called when the program stops / the EventTimer 10 times per second )"
                    Case "load" : s = "Keywords: " + LoadKeywords
                    Case "presskeys" : s = "Keywords: " + GetPressKeysNames()
                    Case "sendkeys" : s = "Keywords: " + GetSendKeysNames()
                    Case "print" : s = "NoKeywords: ""TEXT TO PRINT"" (also complex expressions - use "" to include spaces in the text)"
                    Case "return" : s = "OK: The program will return to the line after ""GoSub"", ""Button"" or ""Key"" events"
                    Case "save" : s = "Keywords: Slots / Vars / StringToFile / DeleteFile"
                    Case "select" : s = "NoKeywords: EXPRESSION"
                    Case "case" : s = "NoKeywords: Please write ""CaseElse"" or an EXPRESSION"
                    Case "caseelse" : s = "OK: Following instructions are executed if no precedent ""Case"" has matched the ""Select"" value"
                    Case "slot" : s = "EXPRESSION"
                    Case "speed" : s = "NoKeywords: Values 1 to 9 - Normal values are: 5(slow execution) 7(fast execution) 9(only for short loops)"
                    Case "stop" : s = "OK: The program will stop here (execution paused)"
                    Case "end" : s = "OK: The program will stop here (execution terminated)"
                    Case "variable" : s = "Keywords: Numeric / String"
                    Case "wait" : s = "Keywords: Seconds / Button / Application / Condition"
                    Case "window" : s = "Keywords: MinSize / Sizable / Maximized / FullScreen"
                    Case Else
                        If IsNumericVariable(sa(0)) OrElse IsStringVariable(sa(0)) Then
                            s = "Keywords: ="
                        Else
                            s = "Keywords: " + AllKeywords
                        End If
                End Select
            Case 2
                Select Case sa(0)
                    Case "button"
                        If IsValidButtonNumber(sa(1)) Or IsValidDouble(sa(1)) Then
                            s = "Keywords: Slot / Text / Color / Enabled / Disabled"
                        Else
                            s = "ERROR: Invalid button number"
                        End If
                    Case "com"
                        If sa(1) = "close" Then
                            s = "OK: This line will close the COM-Port."
                        ElseIf sa(1) = "discardinbuffer" Then
                            s = "OK: This line will discard the COM-Port input buffer."
                        ElseIf sa(1) = "discardoutbuffer" Then
                            s = "OK: This line will discard the COM-Port output buffer."
                        ElseIf sa(1) = "enabledtr" Then
                            s = "OK: This line will enable the DTR output."
                        ElseIf sa(1) = "enablerts" Then
                            s = "OK: This line will enable the RTS output."
                        ElseIf sa(1) = "disabledtr" Then
                            s = "OK: This line will disable the DTR output."
                        ElseIf sa(1) = "disablerts" Then
                            s = "OK: This line will disable the RTS output."
                        Else
                            s = "Keywords: Open / Close / WriteString / WriteLine / DiscardOutBuffer / DiscardInBuffer / EnableDTR / EnableRTS / DisableDTR / DisableRTS"
                        End If
                    Case "controls"
                        If sa(1) = "opentextbox" Then
                            s = "OK: This line will open the print text box"
                        ElseIf sa(1) = "closetextbox" Then
                            s = "OK: This line will close the print text box"
                        ElseIf sa(1) = "cleartextbox" Then
                            s = "OK: This line will clear the print text box"
                        Else
                            s = "Keywords: OpenTextBox / CloseTextBox / ClearTextBox"
                        End If
                    Case "for"
                        If IsNumericVariable(sa(1)) Then
                            s = "Keywords: ="
                        Else
                            s = "ERROR: Invalid control variable"
                        End If
                    Case "gosub"
                        If FindLabel(RTBLines, sa(1)) < 0 Then
                            If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                s = "WARNING: Label " + Quoted(saCase(1)) + " not found. The destination could be wrong at run-time."
                            Else
                                s = "ERROR: Label " + Quoted(saCase(1)) + " not found."
                            End If
                        Else
                            s = "OK: The program will call the subprogram starting with the label " + Quoted(saCase(1))
                        End If
                    Case "goto"
                        If FindLabel(RTBLines, sa(1)) < 0 Then
                            If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                s = "WARNING: Label " + Quoted(saCase(1)) + " not found. The destination could be wrong at run-time."
                            Else
                                s = "ERROR: Label " + Quoted(saCase(1)) + " not found."
                            End If
                        Else
                            s = "OK: The program will jump to the label " + Quoted(saCase(1))
                        End If
                    Case "if"
                        If IsValidBoolean(sa(1)) Then
                            If TestElseOrEndif(nline) Then
                                s = "OK: the expression """ + saCase(1) + """ will be evalued"
                            Else
                                s = "WARNING: This line has no effect without a corresponding Else or EndIf"
                            End If
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "key"
                        If IsValidKey(sa(1)) Then
                            s = "Keywords: Goto / Gosub"
                        Else
                            s = "NoKeywords: Please use a valid Key: 0..9 / a..z / F1..F12 / Left / Right / Up / Down / Pageup / Pagedown / Space / Esc / Home / End"
                        End If
                    Case "label"
                        Dim n As Int32 = FindLabel(RTBLines, sa(1))
                        If n <> nline Then
                            If n >= 0 Then
                                s = "ERROR: The label """ + saCase(1) + """ is already declared."
                            Else
                                s = "ERROR: Expressions are not allowed."
                            End If
                        Else
                            If sa(1) = "eventstop" Then
                                If TestTheFinalEndForTheEventStopLabel(nline) Then
                                    s = "OK: This label will be called when the program stops."
                                Else
                                    s = "ERROR: This structure must be closed with ""End"""
                                End If
                            ElseIf sa(1) = "eventtimer" Then
                                Form1.EventTimer_Init()
                                If Form1.EventTimer_StartLine >= 0 Then
                                    s = "OK: This label will be called ten times per second." + vbCrLf + _
                                        "WARNING: Keep the EventTimer structure very light and fast."
                                Else
                                    s = "ERROR: The EventTimer structure must be closed with Return."
                                End If
                            Else
                                s = "OK: This label will be called by ""GoSub"", ""GoTo"" and by ""Buttons"" with text """ + saCase(1) + """"
                            End If
                        End If
                    Case "load"
                        If sa(1) = "slots" Then
                            s = "OK: All the slots (0 to 999) will be loaded from the ""_Slots.txt"" file."
                        ElseIf sa(1) = "vars" Then
                            s = "OK: All the vars will be loaded from the ""_Vars.txt"" file."
                        ElseIf sa(1) = "pause" Then
                            s = "OK: Sounds and video will be paused here"
                        ElseIf sa(1) = "play" Then
                            s = "OK: Sounds and video will play (if paused)"
                        ElseIf sa(1).StartsWith("position") Then
                            Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(1)).Split(" "c)
                            If sa2.Length = 2 Then
                                s = "OK: Sounds and video will be moved to this position in seconds"
                            Else
                                s = "ERROR: Wrong instruction." + vbCrLf + _
                                    "SUGGESTION: Add position in seconds and fractions."
                            End If
                        ElseIf sa(1) = "stop" Then
                            s = "OK: Sounds and video will be stopped here"
                        ElseIf sa(1) = "hide" Then
                            s = "OK: The running program will be visualized until a new image or video will be loaded."
                        ElseIf sa(1) = "closeapps" Then
                            s = "OK: This line will close all the opened ""Theremino"" applications."
                        ElseIf sa(1).StartsWith("slot") Then
                            Dim n() As Double = ExtractNumericParams(sa(1), 3)
                            If n.Length > 0 Then
                                If n.Length = 2 Then
                                    s = "OK: Slots in the range will be loaded."
                                Else
                                    s = "ERROR: Please write two numbers from 0 to 999, or two expressions without spaces." + vbCrLf + _
                                        "               The first and second values are the first and last Slots to be loaded."
                                End If
                            Else
                                s = "ERROR: Too many parameters."
                            End If
                        Else
                            sa(1) = sa(1).Replace("""", "")
                            Dim str As String = sa(1)
                            Select Case Strings.Right(str, 4).ToLower
                                Case ".txt"

                                    ' --------------------------------------------- test load file
                                    Dim sa2() As String = sa(1).Split(" "c)
                                    If sa2.Length > 1 Then
                                        If IsStringVariable(sa2(0)) Then
                                            s = "OK: The variable """ + sa2(0) + """ will be loaded from the file """ + sa2(sa2.Length - 1) + """"
                                        Else
                                            Dim s2 As String = EvalString_AutoQuoting(sa(1))
                                            If IsValidString_AutoQuoting(s2) Then
                                                s = "OK: The program """ + sa(1) + """ will be loaded from ""Programs"" folder"
                                            Else
                                                s = "ERROR: The variable """ + sa2(0) + """ is not declared"
                                            End If
                                        End If
                                    End If

                                Case ".jpg", ".bmp", ".png", ".gif" : s = "OK: The image will be loaded from ""Media"" folder"
                                Case ".avi", ".wmv", ".mpg", ".mp4", ".mov" : s = "OK: The video will be loaded from ""Media"" folder"
                                Case ".mp3", ".wav" : s = "OK: The audio file will be played (with the video player)"
                                Case ".exe" : s = "OK: The application will be started (if exists near to file 'Automation.exe' or in the 'Full-Path' specified)"
                                Case Else
                                    If IsStringVariable(sa(1)) Then
                                        s = "OK: The image, video or program, will be loaded from ""Media"" or ""Programs"" folder"
                                    Else
                                        If str.StartsWith("http") Then
                                            s = "OK: The site will be opened (if exists)"
                                        Else
                                            s = "OK: Will try to load the """ + str + """ (if exists)" + vbCrLf
                                            s += "Suggestions: Slots / Vars / Stop / Pause / Play / Position / Hide / CloseApps"
                                        End If
                                    End If
                            End Select
                        End If
                    Case "presskeys"
                        s = "Keywords: " + GetPressKeysNames()
                    Case "sendkeys"
                        s = "Keywords: " + GetSendKeysNames()
                    Case "print"
                        If IsValidBooleanStringOrDouble(sa(1)) Then
                            s = "OK: The expression """ + saCase(1) + """ will be evalued and printed"
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "save"
                        If sa(1) = "slots" Then
                            s = "OK: All the slots (0 to 999) will be saved to the ""_Slots.txt"" file."
                        ElseIf sa(1) = "vars" Then
                            s = "OK: All the vars will be saved to the ""_Vars.txt"" file."
                        Else
                            If sa(1).StartsWith("stringtofile") Then
                                s = "NoKeywords: Please write a String and a Filename"
                            Else
                                s = "Keywords: Slots / Vars / StringToFile / DeleteFile"
                            End If
                        End If
                    Case "select"
                        If TestCase(nline) And TestEndselect(nline) Then
                            If IsValidBooleanStringOrDouble(sa(1)) Then
                                s = "OK: The expression """ + saCase(1) + """ will be compared with ""Case"" expressions"
                            Else
                                s = "NoKeywords: Please write a valid expression"
                            End If
                        Else
                            s = "WARNING: This line has no effect without corresponding Case and EndSelect"
                        End If
                    Case "case"
                        If TestEndselect(nline) Then
                            If IsValidBooleanStringOrDouble(sa(1)) Then
                                s = "OK: Following instructions are executed when the ""Case"" value equals the ""Select"" value"
                            Else
                                s = "NoKeywords: Please write a valid expression"
                            End If
                        Else
                            s = "WARNING: This line has no effect without a corresponding EndSelect"
                        End If
                    Case "slot"
                        If IsValidDouble(sa(1)) Then
                            s = "Keywords: ="
                        Else
                            s = "NoKeywords: Please write a valid numerical expression"
                        End If
                    Case "speed"
                        If IsValidDouble(sa(1)) Then
                            s = "OK: The speed will be set to """ + saCase(1) + """"
                        Else
                            s = "NoKeywords: Please write 1 to 9 or a valid numeric expression"
                        End If
                    Case "variable"
                        Select Case sa(1)
                            Case "numeric", "string"
                                s = "NoKeywords: Please write a name for this variable"
                            Case Else
                                s = "Keywords: Numeric / String"
                        End Select
                    Case "wait"
                        Select Case sa(1)
                            Case "seconds" : s = "NoKeywords: Please write a time in seconds (also with decimals)"
                            Case "button"
                                s = "NoKeywords: BUTTON-NAME"
                            Case "application"
                                If IsValidString(sa(1)) Then
                                    s = "NoKeywords: APP-NAME"
                                Else
                                    s = "ERROR: Invalid app-name."
                                End If
                            Case Else
                                If IsValidBoolean(sa(1)) Then
                                    s = "OK: This line waits until the expression """ + saCase(1) + """ will be TRUE"
                                Else
                                    s = "Keywords: Seconds / Button / Key / Application / Condition"
                                End If
                        End Select
                    Case "window"
                        If sa(1) = "minsize" OrElse _
                           sa(1) = "sizable" OrElse _
                           sa(1) = "maximized" OrElse _
                           sa(1) = "fullscreen" Then
                            s = "OK: The window will be set to " + saCase(1)
                        Else
                            s = "Keywords: MinSize / Sizable / Maximized / FullScreen"
                        End If
                    Case Else
                        If sa(1) = "=" Then
                            If IsNumericVariable(sa(0)) Then
                                s = "NoKeywords: Please write a valid numerical expression, or the keyword: Input"
                            ElseIf IsStringVariable(sa(0)) Then
                                s = "NoKeywords: Please write a valid string (with "" initial and final), or the keyword: Input"
                            Else
                                s = "ERROR: " + LastError
                            End If
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                End Select
            Case 3
                Select Case sa(0)
                    Case "button"
                        If IsValidButtonNumber(sa(1)) Or IsValidDouble(sa(1)) Then
                            Select Case sa(2)
                                Case "enabled", "disabled"
                                    s = "OK: Button will be " + saCase(2)
                                Case "slot" : s = "NoKeywords: SLOT-NUMBER"
                                Case "text" : s = "NoKeywords: BUTTON-TEXT  (enclosing the text in double quotes is possible, but not necessary)"
                                Case "color"
                                    s = "NoKeywords: Please write ""Flashing"" or a color name like ""Yellow"", ""Blue"", ""Red""..."
                                    SuggestionsButtonType = "Colors"
                                Case Else : s = "Keywords: Slot / Text / Color / Enabled / Disabled"
                            End Select
                        Else
                            s = "ERROR: Invalid button number"
                        End If
                    Case "com"
                        Select Case sa(1)
                            Case "writestring"
                                If IsValidString_AutoQuoting(sa(2)) Then s = "OK: The string " + saCase(2) + " will be written on the serial line"
                            Case "writeline"
                                If IsValidString_AutoQuoting(sa(2)) Then s = "OK: The string " + saCase(2) + " will be written on the serial line, followed by a final CRLF"
                            Case Else : s = "Keywords: Open / WriteString / WriteLine"
                        End Select
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" Then
                            s = "NoKeywords: Please write a numeric expression (without spaces), or a numeric variable."
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "key"
                        If IsValidKey(sa(1)) Then
                            If sa(2) = "goto" OrElse sa(2) = "gosub" Then
                                s = "NoKeywords: Please write a Label name   (enclosing the text in double quotes is possible, but not necessary)"
                            Else
                                s = "Keywords: Goto / Gosub"
                            End If
                        Else
                            s = "ERROR: The key is not valid"
                        End If
                    Case "save"
                        If sa(1) = "deletefile" And IsValidString_AutoQuoting(sa(2)) Then
                            s = "OK: The file " + Quoted(saCase(2).Trim) + " will be deleted"
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "slot"
                        If IsValidDouble(sa(1)) Then
                            If sa(2) = "=" Then
                                s = "NoKeywords: Please write RESET, SLEEP or a valid numerical expression"
                            Else
                                s = "NoKeywords: Please write a ""="""
                            End If
                        Else
                            s = "ERROR: The slot number is not valid"
                        End If
                    Case "wait"
                        Select Case sa(1)
                            Case "seconds"
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: The program will be paused for """ + saCase(2) + """ seconds"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case "button"
                                s = "OK: The program will wait the button """ + saCase(2) + """ if exists. (for buttons with long names the first word is used)"
                            Case "application"
                                If IsValidString(sa(2)) Then
                                    s = "OK: The program will wait the application """ + EvalString(sa(2)) + """"
                                Else
                                    s = "ERROR: Invalid application name."
                                End If
                            Case Else
                                s = "OK: The expression """ + saCase(2) + """ will be evaluated."
                        End Select
                    Case "variable"
                        Select Case sa(1)
                            Case "numeric"
                                ' ------------------------------------- suggestion: OK or ERROR
                                If IsNumericUserVariable(sa(2)) Then
                                    If IsNumericUserVariable(sa(2), nline) Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: The variable """ + saCase(2) + """ is already declared."
                                    End If
                                Else
                                    s = TestUserVariableName(sa(2))
                                    If s = "OK" Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: " + s
                                    End If
                                End If
                            Case "string"
                                ' ------------------------------------- suggestion: OK or ERROR 
                                If IsStringUserVariable(sa(2)) Then
                                    If IsStringUserVariable(sa(2), nline) Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: The variable """ + saCase(2) + """ is already declared."
                                    End If
                                Else
                                    s = TestUserVariableName(saCase(2))
                                    If s = "OK" Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: " + s
                                    End If
                                End If
                            Case Else
                                s = "ERROR: Please correct with ""Numeric"" or ""String"""
                        End Select
                    Case Else
                        If sa(1) = "=" Then
                            If IsNumericVariable(sa(0)) Then
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: Numeric variable """ + saCase(0) + """ = " + saCase(2)
                                ElseIf IsValidBoolean(sa(2)) Then
                                    s = "OK: Boolean expression """ + saCase(0) + """ = " + saCase(2)
                                ElseIf sa(2) = "input" Then
                                    s = "OK: Numeric variable """ + saCase(0) + """ will be manually entered."
                                ElseIf sa(2).StartsWith("i") Then
                                    s = "Keywords: Input"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            ElseIf IsStringVariable(sa(0)) Then
                                If IsValidString(sa(2)) Then
                                    s = EvalString(saCase(2))
                                    If s = "" Then s = saCase(2)
                                    s = "OK: String variable """ + saCase(0) + """ = " + Quoted(s)
                                ElseIf sa(2) = "input" Then
                                    s = "OK: String variable """ + saCase(0) + """ will be manually entered."
                                ElseIf sa(2).StartsWith("i") Then
                                    s = "Keywords: Input"
                                Else
                                    s = "ERROR: Invalid string. " + LastError
                                End If
                            Else
                                s = "ERROR: The variable """ + saCase(0) + """ is not declared."
                            End If
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                End Select
            Case 4
                Select Case sa(0)
                    Case "button"
                        If Not IsValidButtonNumber(sa(1)) And Not IsValidDouble(sa(1)) Then
                            s = "ERROR: Invalid button number."
                        Else
                            Select Case sa(2)
                                Case "slot"
                                    s = "OK: Button """ + saCase(1) + """ will be pressed by slot """ + saCase(3) + """ (when value exceeds 500)"
                                Case "text"
                                    If FindLabel(RTBLines, sa(3)) < 0 Then
                                        If sa(3).Contains(" ") Or IsSomeVariable(sa(3)) Then
                                            s = "WARNING: Label " + Quoted(saCase(3)) + " not found. The destination could be wrong at run-time."
                                        Else
                                            If sa(3).Replace("""", "").Length > 0 Then
                                                s = "WARNING: Label " + Quoted(saCase(3)) + " not found."
                                            Else
                                                s = "OK: This instruction will hide the button """ + saCase(1) + """"
                                            End If
                                        End If
                                    Else
                                        s = "OK: Button """ + saCase(1) + """ text is """ + saCase(3) + """" + vbCrLf + "        (the first word """ + ExtractFirstWord(saCase(3)) + """ will be used for ""Gosub Labels"" and ""Wait Button"")"
                                    End If
                                Case "color"
                                    Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(3)).Split(" "c)
                                    If sa(3).StartsWith("flashing") Then
                                        If sa2.Length = 1 Then
                                            s = "Keywords: " + GetColorNames.Replace(" ", " / ")
                                            SuggestionsButtonType = "Colors"
                                        ElseIf sa2.Length = 2 Then
                                            s = "Keywords: " + GetColorNames.Replace(" ", " / ")
                                            SuggestionsButtonType = "Colors"
                                        ElseIf sa2.Length = 3 Then
                                            If ColorFromColorName(sa2(1)) = Color.Empty Or _
                                               ColorFromColorName(sa2(2)) = Color.Empty Then
                                                s = "Keywords: " + GetColorNames.Replace(" ", " / ")
                                                SuggestionsButtonType = "Colors"
                                            Else
                                                s = "OK: The button """ + saCase(1) + """ background will be set with the flashing colors: """ + saCase(3) + """" + vbCrLf + _
                                                    "(use ""Flashing1"" to ""Flashing9"" to change the flashing speed)"
                                            End If
                                        Else
                                            s = "ERROR: Incomplete or wrong instruction. "
                                        End If
                                    Else
                                        If sa2.Length < 2 Then
                                            If ColorFromColorName(sa(3)) = Color.Empty Then
                                                s = "Keywords: Flashing / " + GetColorNames.Replace(" ", " / ")
                                                SuggestionsButtonType = "Colors"
                                            Else
                                                s = "OK: The button """ + saCase(1) + """ background will be set with the color: """ + saCase(3) + """"
                                            End If
                                        Else
                                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                                        End If
                                    End If
                                Case Else
                                    s = "ERROR: Incomplete or wrong instruction. " + LastError
                            End Select
                        End If
                    Case "com"
                        If sa(1) = "open" Then
                            If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                                s = "OK: The COM port " + sa(2) + " will be opened with Bauds " + sa(3)
                            Else
                                s = "ERROR: " + LastError
                            End If
                        End If
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) Then
                            s = "Keywords: To"
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                    Case "key"
                        If IsValidKey(sa(1)) Then
                            If FindLabel(RTBLines, sa(3)) < 0 Then
                                If sa(3).Contains(" ") Or IsSomeVariable(sa(3)) Then
                                    s = "WARNING: Label " + Quoted(saCase(3)) + " not found. The destination could be wrong at run-time."
                                Else
                                    s = "ERROR: Label " + Quoted(saCase(3)) + " not found."
                                End If
                            Else
                                If sa(2) = "goto" Then
                                    s = "OK: Pressing the key """ + saCase(1) + """ the Label """ + saCase(3) + """ will be called by a GoTo"
                                ElseIf sa(2) = "gosub" Then
                                    s = "OK: Pressing the key """ + saCase(1) + """ the Label """ + saCase(3) + """ will be called by a GoSub (remember to use a Return)"
                                Else
                                    s = "Keywords: Goto / Gosub"
                                End If
                            End If
                        Else
                            s = "ERROR: The key is not valid"
                        End If
                    Case "load"
                        If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                            s = "OK: This statement will load Slots from " + saCase(2).Trim + " to " + saCase(3).Trim
                        Else
                            s = "ERROR: Please write two simple numeric values from 0 to 999, or variables V1 to V9 " + vbCrLf + _
                                "                (the first value is first Slot and the second value is the last Slot to load)"
                        End If
                    Case "save"
                        If sa(1) = "stringtofile" And IsValidString_AutoQuoting(sa(2)) And IsValidString_AutoQuoting(sa(3)) Then
                            s = "OK: The string " + Quoted(saCase(2).Trim) + " will be appended to the file: " + Quoted(saCase(3).Trim)
                        Else
                            s = "ERROR: Please write a string to write and a string with the filename"
                        End If
                    Case "slot"
                        If IsValidDouble(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) Then
                            s = "OK: Slot """ + saCase(1).Trim + """ = expression: """ + saCase(3).Trim + """"
                        Else
                            Select Case sa(3)
                                Case "reset", "sleep"
                                    s = "OK: Slot """ + saCase(1).Trim + """ will be set with the special value: NAN+" + saCase(3)
                                Case Else
                                    s = "ERROR: Please write RESET , SLEEP or a valid numeric expression"
                            End Select
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
                Select Case sa(2)
                    Case "input"
                        s = "OK: The variable """ + saCase(0) + """ will be manually entered." + vbCrLf + _
                            "         (after the keyword ""Input"" you could also write a prompt for the user)"
                End Select

            Case 5
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) And sa(4) = "to" Then
                            s = "NoKeywords: Please write a numeric expression (without spaces), or a numeric variable."
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case 6
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) Then
                            If sa(2) = "=" And IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) Then
                                If TestForNext(nline) Then
                                    s = "OK: Initially the variable """ + sa(1) + """ is set with the """ + sa(3) + """ numeric value." + vbCrLf
                                    s += "       Then the lines between For and Next will be executed until """ + sa(1) + """ reaches or exceeds """ + sa(5) + """"
                                Else
                                    s = "ERROR: This FOR has not the closing NEXT"
                                End If
                            Else
                                s = "ERROR: " + LastError
                            End If
                        Else
                            s = "ERROR: The control variable """ + sa(1) + """ is not declared (or not numeric)."
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case 7
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                           IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) Then
                            If sa(6) = "step" Then
                                s = "OK: Please write a step value"
                            Else
                                s = "OK: Please write the word ""STEP"", and a step value"
                            End If
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case 8
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                           IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) And _
                           sa(6) = "step" And IsValidDouble(sa(7)) Then
                            If TestForNext(nline) Then
                                s = "OK: Initially the variable """ + sa(1) + """ is set with the """ + sa(3) + """ numeric value." + vbCrLf
                                s += "       Then the variable """ + sa(1) + """ will be incremented by """ + sa(7) + """ until reaches or exceeds """ + sa(5) + """"
                            Else
                                s = "ERROR: This FOR has not the closing NEXT"
                            End If
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case Else
                s = "ERROR: Incomplete or wrong instruction. " + LastError
        End Select
        ' -----------------------------------------------------------
        Form1.SetStatusLabelFontProportional()
        With Form1.StatusLabel1
            ' ------------------------------------------------------- open-close the Suggestions Button 
            Select Case SuggestionsButtonType
                Case "Colors"
                    .Width = Form1.StatusStrip1.Width - Form1.chk_Suggestions.Width - 28
                    Form1.chk_Suggestions.Text = "Color selector"
                    Form1.chk_Suggestions.Visible = True
                Case Else
                    .Width = Form1.StatusStrip1.Width - 22
                    Form1.chk_Suggestions.Visible = False
            End Select
            ' ------------------------------------------------------- status label colors
            If s.StartsWith("OK") Then
                If s.ToLower.Contains("suggestions:") Then
                    .BackColor = Color.FromArgb(210, 255, 150)
                Else
                    .BackColor = Color.FromArgb(250, 255, 200)
                End If
                .Text = s
            ElseIf s.ToLower.StartsWith("error") Then
                If s.Contains("Suggestions:") Then
                    .BackColor = Color.FromArgb(210, 255, 150)
                Else
                    .BackColor = Color.FromArgb(255, 160, 150)
                End If
                .Text = s
            ElseIf s.ToLower.StartsWith("warning") Then
                .BackColor = Color.FromArgb(255, 200, 100)
                .Text = s
            ElseIf s = "" Then
                .BackColor = Color.AliceBlue
                .Text = ""
            Else
                If sa.Length > 0 Then
                    ' ----------------------------------------------- Keyword auto selection
                    Dim s2 As String
                    If s.ToLower.StartsWith("nokeywords: ") Then
                        s2 = s
                    Else
                        s2 = ExtractKeywords(s, SelectedWord)
                    End If
                    If s2 <> "" Then
                        s = s2
                        If s2.ToLower.StartsWith("nokeywords: ") Then
                            s = s.Substring(12)
                        Else
                            If Not s.ToLower.Trim.StartsWith("keywords") Then
                                s = "Keywords: " + s
                            End If
                        End If
                    Else
                        If sa(0) IsNot Nothing AndAlso _
                                 (sa(0).ToLower = "sendkeys" Or sa(0).ToLower = "presskeys") Then
                            s = "OK: The statement is valid"
                        Else
                            .BackColor = Color.FromArgb(255, 160, 150)
                            .Text = "Error:  Incomplete or wrong instruction. " + LastError
                            Return
                        End If
                    End If
                End If
                ' --------------------------------------------------- Keyword suggestions
                .BackColor = Color.FromArgb(220, 255, 200)
                .Text = s
            End If
        End With
    End Sub

    Private Function Quoted(ByVal s As String) As String
        If Not s.StartsWith("""") And Not s.EndsWith("""") Then
            Return """" + s + """"
        End If
        Return s
    End Function

    Private Function UnusedVariable(ByVal s As String, ByVal nline As Int32) As Boolean
        Dim l As String
        ' ------------------------------------------------ add spaces before and after 
        s = " " + s + " "
        ' ------------------------------------------------ test all the lines
        For i As Int32 = 0 To RTBLines.Length - 1
            If i = nline Then Continue For
            l = RTBLines(i).Trim.ToLower
            ' -------------------------------------------- string with spaces around operators
            l = StringSpacedAroundOperators(l)
            ' -------------------------------------------- 
            If l.Contains(s) Then
                If l.StartsWith("beep ") Then Continue For
                If l.StartsWith("key ") Then Continue For
                If l.StartsWith("label ") Then Continue For
                If l.StartsWith("variable ") Then Continue For
                If l.IndexOf("'") >= 0 AndAlso l.IndexOf(s) > l.IndexOf("'") Then Continue For
                Return False
            End If
        Next
        Return True
    End Function



    ' ===============================================================================================
    '  AUTOMATIC COMPLETION
    ' ===============================================================================================
    Private Function ExtractKeywords(ByVal s1 As String, ByVal s2 As String) As String
        If s2 = "" Then Return s1
        ExtractKeywords = s1
        Dim MatchesCount As Int32 = 0
        s1 = s1.Replace("Keywords:", "")
        Dim s() As String = Split(s1.Replace(" / ", " "), " ")
        s1 = ""
        For Each word As String In s
            If word.StartsWith(s2, StringComparison.InvariantCultureIgnoreCase) Then
                If s1 <> "" Then s1 += " / "
                s1 += word
            End If
        Next
        Return s1
    End Function

    Private Sub CompleteKeyword(ByVal keyword As String, ByVal AddSpace As Boolean)
        ' --------------------------------------------------------------- save
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        PaintEventEnabled = False
        ' No "Enabled = false" here otherwise the scroll bar flashes
        Dim OldCursorPosition As Integer = SelectionStart
        ' --------------------------------------------------------------- change
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        SelectionStart = start
        SelectionLength = length
        If AddSpace Then
            SelectedText = keyword & " "
        Else
            SelectedText = keyword
        End If
        ' --------------------------------------------------------------- restore
        SelectionLength = 0
        PaintEventEnabled = True
        EventsAreEnabled = old
        ' --------------------------------------------------------------- update suggestions
        SelectionStart = SelectionStart - 1
        SelectionStart = SelectionStart + 1
    End Sub

    Friend Sub CompleteKeyword()
        Dim s As String = Form1.StatusLabel1.Text.Trim
        If s.Contains("/") Then Return
        If s.ToLower.StartsWith("keywords") Then
            Dim sa() As String
            sa = s.Split(" "c)
            If sa.Length > 1 And sa.Length < 8 Then
                CompleteKeyword(sa(sa.Length - 1).Trim, True)
                ColorSelectedLine()
            End If
        Else
            Dim n As Int32 = s.IndexOf("Suggestions")
            If n > 0 Then
                s = Mid(s, n + 14).Trim
                Dim sa() As String
                sa = s.Split(" "c)
                If sa.Length = 1 Then
                    CompleteKeyword(sa(0).Trim, False)
                    ColorSelectedLine()
                End If
            End If
        End If
    End Sub

    Friend Sub CompleteKeyword_MouseSelected(ByVal str As String)
        CompleteKeyword(str, True)
        ColorSelectedLine()
    End Sub

    ' ===============================================================================================
    '  CUT COPY and PASTE
    ' ===============================================================================================
    Friend Sub ExecCut()
        If Me.SelectionLength > 0 Then
            TestKeywords_PrepareKeywordsToRecolor()
            SendTextToClipboard()
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            Me.SelectedText = ""
            ' ---------------------------------------------------------------- 
            ShowSuggestions()
            ReinitUserVariables()
            TestKeywords_RecolorLines()
            ColorAllLines_StartingWithVariable()
        End If
    End Sub
    Friend Sub ExecCopy()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
        End If
    End Sub
    Private Sub SendTextToClipboard()
        Dim data As IDataObject
        data = New DataObject(DataFormats.Text, Me.SelectedText.Replace(vbLf, vbCrLf))
        Clipboard.SetDataObject(data, True)
    End Sub
    Friend Sub ExecPaste()
        If Clipboard.GetDataObject().GetDataPresent(DataFormats.Text) Then
            ' ------------------------------------------------------------ 
            TestKeywords_PrepareKeywordsToRecolor()
            ' ------------------------------------------------------------ 
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------ extract only the text from clipbbard
            Dim TextToPaste As String = Clipboard.GetText
            '
            ' ------------------------------------------------------------ prepare StartLine and EndPosition
            Dim StartLine As Int32 = GetLineFromCharIndex(SelectionStart)
            Dim EndPosition As Int32 = SelectionStart + SelectionLength - SelectedText.Length + TextToPaste.Length
            '
            ' ------------------------------------------------------------ paste the text and prepare EndLine
            Me.SelectedText = TextToPaste
            Dim EndLine As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            '
            ' ------------------------------------------------------------ color lines
            ReinitUserVariables()
            ColorLines(StartLine, EndLine)
            ColorAllLines_StartingWithVariable()
            ' ------------------------------------------------------------
            ShowSuggestions()
            ' ------------------------------------------------------------
            Me.SelectLines(StartLine, EndLine)
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
            '
            ' ------------------------------------------------------------ set the final cursor position
            Me.SelectionLength = 0
            If TextToPaste.Contains(vbLf) Then
                Me.SelectionStart = Me.GetLastCharIndexFromLine(EndLine)
            Else
                Me.SelectionStart = EndPosition
            End If
            '
            ' ------------------------------------------------------------
            UndoAndHistory_AddEntry()
            Me.Enabled = True
            PaintEventEnabled = True
            Me.Focus()
        End If
    End Sub
    Friend Sub ExecDelete()
        If Me.SelectionLength > 0 Then
            ' ------------------------------------------------------------ 
            TestKeywords_PrepareKeywordsToRecolor()
            ' ------------------------------------------------------------ 
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------ 
            Dim startline As Int32 = GetLineFromCharIndex(SelectionStart)
            Me.SelectedText = ""
            Dim endline As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            ' ------------------------------------------------------------ color lines
            ReinitUserVariables()
            ColorLines(startline, endline)
            ColorAllLines_StartingWithVariable()
            ' ------------------------------------------------------------
            ShowSuggestions()
            ' ------------------------------------------------------------
            Me.SelectLines(startline, endline)
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
            Me.SelectionLength = 0
            Me.SelectionStart = Me.GetLastCharIndexFromLine(endline)
            ' ------------------------------------------------------------
            UndoAndHistory_AddEntry()
            Me.Enabled = True
            PaintEventEnabled = True
            Me.Focus()
        End If
    End Sub
    'Friend Function Editor_CanPaste() As Boolean
    '    Return Clipboard.GetDataObject().GetDataPresent(DataFormats.Text)
    'End Function

    ' ===============================================================================================
    '  BREAKPOINTS and WATCHES
    ' ===============================================================================================
    Friend Sub ToggleBreakPoint(ByVal mousepos As Point)
        Dim charindex As Int32 = Me.GetCharIndexFromPosition(mousepos)
        Dim rightpos As Point = Me.GetPositionFromCharIndex(charindex)
        Dim nline As Int32 = GetLineFromCharIndex(charindex)
        Dim lineText As String = ""
        If RTBLines.Length > 0 Then lineText = RTBLines(nline).TrimEnd
        If lineText.EndsWith("'Breakpoint") Then
            lineText = lineText.Replace("'Breakpoint", "").TrimEnd
            If charindex >= GetFirstCharIndexFromLine(nline) + lineText.Length Then
                DeleteBreakpointFromLine(nline)
            End If
        Else
            If mousepos.X > rightpos.X + 8 Then
                ' ---------------------------------------------------------------------
                Dim len As Int32 = GetLineLengthAsEquivalentSpaces(nline)
                Dim ntabs As Int32 = (48 - len + 3) \ 4
                If ntabs < 0 Then ntabs = 0
                ' ---------------------------------------------------------------------
                Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(nline) + RTBLines(nline).Length
                Me.Select(endpos, 0)
                Me.SelectedText = StrDup(ntabs, vbTab) + " 'Breakpoint"
            End If
        End If
    End Sub

    Friend Sub DeleteAllBreakPoints()
        Dim i As Int32 = RTBLines.Length - 1
        While i >= 0
            Dim lineText As String = RTBLines(i).Trim
            If lineText.EndsWith("'Breakpoint") Then
                DeleteBreakpointFromLine(i)
            Else
                i = i - 1
            End If
        End While
    End Sub

    Private Sub DeleteBreakpointFromLine(ByVal i As Int32)
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTBLines(i).Length
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTBLines(i).IndexOf("'Breakpoint")
        Me.Select(startpos, endpos - startpos)
        Me.SelectedText = ""
        ' --------------------------------------------- TrimRightSpacesFromLine
        TrimRightSpacesFromLine_TrimAlsoEmptyLines(i)
        ' --------------------------------------------- restore tabs and position the cursor after them
        AutoIndent()
        Me.Select(Me.GetFirstCharIndexFromLine(i) + RTBLines(i).Length, 0)
    End Sub

    Private Sub TrimRightSpacesFromLine_TrimAlsoEmptyLines(ByVal i As Int32)
        If i < 0 Or i >= RTBLines.Length Then Return
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTBLines(i).Length
        RTBLines(i) = RTBLines(i).TrimEnd
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTBLines(i).Length
        If endpos > startpos Then
            Me.Select(startpos, endpos - startpos)
            Me.SelectedText = ""
            RTBLines = Me.Lines
        End If
    End Sub

    Private Sub TrimRightSpacesFromLine(ByVal i As Int32)
        If i < 0 Or i >= RTBLines.Length Then Return
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTBLines(i).Length
        If RTBLines(i).Trim = "" Then
            While RTBLines(i).EndsWith(" ")
                RTBLines(i) = RTBLines(i).Remove(RTBLines(i).Length - 1, 1)
            End While
        Else
            RTBLines(i) = RTBLines(i).TrimEnd
        End If
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTBLines(i).Length
        If endpos > startpos Then
            Me.Select(startpos, endpos - startpos)
            Me.SelectedText = ""
            RTBLines = Me.Lines
        End If
    End Sub

    Friend Sub GoToNextOccurrence()
        ' ---------------------------------------------------------- get the word to be searched
        Dim s As String = Me.SelectedText
        If s = "" Then
            s = GetCompleteWord_Short(Me.Text, Me.SelectionStart)
        End If
        s = s.ToLower.Replace("""", "").Replace("'", "")
        If s = "" Then Return
        ' ---------------------------------------------------------- search all the lines starting from cursor
        Dim l As String
        Dim startLine As Int32 = Me.GetLineFromCharIndex(Me.SelectionStart)
        For testLine As Int32 = startLine + 1 To 99999999
            If testLine >= RTBLines.Length - 1 Then testLine = 0
            If testLine = startLine Then Exit For
            l = RTBLines(testLine).ToLower.Trim.Replace("""", "")
            l = l.Replace("'", "")
            l = l.TrimEnd
            If l.Contains(s) Then
                Me.ScrollToCenterSelectedLine(testLine)
                Dim n As Int32 = RTBLines(testLine).ToLower.IndexOf(s)
                Me.Select(GetFirstCharIndexFromLine(testLine) + n, s.Length)
                History_AddEntry()
                Me.Focus()
                Exit For
            End If
        Next
    End Sub

    Friend Sub GoToDeclaration()
        ' ---------------------------------------------------------- get the word to be searched
        Dim s As String = Me.SelectedText
        If s = "" Then
            s = GetCompleteWord_Short(Me.Text, Me.SelectionStart)
        End If
        s = s.ToLower.Replace("""", "").Replace("'", "")
        If s = "" Then Return
        ' ---------------------------------------------------------- search all the lines starting from cursor
        Dim l As String
        Dim startLine As Int32 = Me.GetLineFromCharIndex(Me.SelectionStart)
        For testLine As Int32 = startLine + 1 To 99999999
            If testLine >= RTBLines.Length - 1 Then testLine = 0
            If testLine = startLine Then Exit For
            l = RTBLines(testLine).ToLower.Trim.Replace("""", "")
            l = l.Replace("'", "")
            l = l.TrimEnd
            If l.Contains(s) AndAlso _
               (l.StartsWith("label") Or l.StartsWith("variable") Or _
                l.StartsWith("button") Or l.StartsWith("key")) Then
                Me.ScrollToCenterSelectedLine(testLine)
                Dim n As Int32 = RTBLines(testLine).ToLower.IndexOf(s)
                Me.Select(GetFirstCharIndexFromLine(testLine) + n, s.Length)
                History_AddEntry()
                Me.Focus()
                Exit For
            End If
        Next
    End Sub

    Friend Sub LoadProgramFile(ByVal fname As String)
        ' ----------------------------------------------- load as unicode (or any other format specified in the txt)
        Dim s As String() = IO.File.ReadAllLines(fname)
        ' ----------------------------------------------- trim end
        For i As Int32 = 0 To s.Length - 1
            s(i) = s(i).TrimEnd
        Next
        ' ----------------------------------------------- 
        Me.Lines = s
        Me.Text += vbCrLf
        RTBLines = Me.Lines
    End Sub

    Friend Sub ExtractWatches()
        RTBLines = Me.Lines
        Dim startline As Int32 = 0
        ' ------------------------------------------- a maximum of 100 watches to increase the loading speed
        If RTBLines.Length > 101 Then
            startline = RTBLines.Length - 100
        End If
        ' -------------------------------------------
        Form_Debug.DeleteAllWatches()
        Form_Debug.txt_Exec.Text = ""
        Dim i As Int32 = startline
        While i < RTBLines.Length - 1
            Dim lineText As String = RTBLines(i).Trim
            If lineText.StartsWith("'WATCH") Then
                Form_Debug.AddWatchLine(Mid(lineText, 7).Trim)
                DeleteLine(i)
                RTBLines = Me.Lines
                i -= 1
            End If
            If lineText.StartsWith("'EXEC") Then
                Form_Debug.txt_Exec.Text = Mid(lineText, 6).Trim
                DeleteLine(i)
                RTBLines = Me.Lines
                i -= 1
            End If
            i += 1
        End While
        Form_Debug.ResizeListViewColumns()
    End Sub

    Friend Sub DeleteLine(ByVal line As Int32)
        Dim startpos As Int32 = GetFirstCharIndexFromLine(line)
        Dim count As Int32 = Lines(line).Length + 1
        If startpos + count > Me.Text.Length Then Return
        Me.Text = Me.Text.Remove(startpos, count)
    End Sub

    Friend Sub DeleteFinalEmptyLines()
        RTBLines = Me.Lines
        For i As Int32 = RTBLines.Length - 1 To 0 Step -1
            If RTBLines(i).Trim = "" Then
                ReDim Preserve RTBLines(RTBLines.Length - 2)
            Else
                Exit For
            End If
        Next
        Me.Lines = RTBLines
        Me.Text += vbCrLf
    End Sub

    ' ===============================================================================================
    '  INDENT-UNINDENT
    ' ===============================================================================================
    Friend Sub IndentSelectedLines(ByVal increment As Int32)
        If RTBLines.Length > 0 Then
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTBLines(i)
                Dim ntabs As Int32 = CountInitialTabs_NotAllowingSpaces(lineText, 4) + increment
                If ntabs >= 0 Then
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                    Me.Select(start, CountInitialChars(lineText))
                    Me.SelectedText = StrDup(ntabs, vbTab)
                End If
            Next
            ' ----------------------------------------------------------------
            SelectLines(FirstSelectedLine, LastSelectedLine)
            ' ----------------------------------------------------------------
            Me.Enabled = True
            PaintEventEnabled = True
            UndoAndHistory_AddEntry()
        End If
        Me.Focus()
    End Sub

    ' ===============================================================================================
    '  COMMENT UN-COMMENT
    ' ===============================================================================================
    Friend Sub CommentSelectedLines(ByVal comment As Boolean)
        ' ------------------------------------------------------------ 
        TestKeywords_PrepareKeywordsToRecolor()
        ' ------------------------------------------------------------ 
        RTBLines = Me.Lines
        If RTBLines.Length > 0 Then
            ' ----------------------------------------------------------------
            If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim CommentColumn As Int32 = -1
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTBLines(i)
                Dim firstch As Int32 = CountInitialChars(lineText)
                If firstch >= 0 Then
                    If CommentColumn < 0 Then CommentColumn = firstch
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)

                    If comment Then
                        Me.Select(start + CommentColumn, 0)
                        Me.SelectedText = "'"
                    Else
                        Me.Select(start + firstch, 1)
                        If Me.SelectedText = "'" Then
                            Me.SelectedText = ""
                        End If
                    End If
                End If
            Next
            ' ----------------------------------------------------------------
            ReinitUserVariables()
            ColorLines(FirstSelectedLine, LastSelectedLine)
            ColorAllLines_StartingWithVariable()
            ' ----------------------------------------------------------------
            Me.Enabled = True
            PaintEventEnabled = True
            UndoAndHistory_AddEntry()
            ' ------------------------------------------------------------ 
            Me.SelectLines(FirstSelectedLine, LastSelectedLine)
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
        End If
        Me.Focus()
    End Sub

    Private Function CountInitialChars(ByVal s As String) As Int32
        Dim CharCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : CharCount += 1
                Case " " : CharCount += 1
                Case Else : Exit For
            End Select
        Next
        Return CharCount
    End Function

    Private Function CountInitialTabs(ByVal s As String, ByVal TabSize As Int32) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function

    Private Function CountInitialTabs_NotAllowingSpaces(ByVal s As String, ByVal TabSize As Int32) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case " " : Return 9999           ' Spaces are not allowed
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function

    


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '  AUTO COMPLETION - AUTOCOMPLETION
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Private AutoCompletion_FinalCursorPosition As Int32

    Private Function TestAutoCompletion() As Boolean
        AutoCompletion_FinalCursorPosition = 0
        Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
        If FirstSelectedLine < 0 Then Return False
        If FirstSelectedLine >= RTBLines.Length Then Return False
        ' --------------------------------------------------------
        Dim s As String = RTBLines(FirstSelectedLine).Trim.ToLower
        Dim status As String = Form1.StatusLabel1.Text.ToLower
        If s.StartsWith("if") Then
            If status.StartsWith("warning") And status.EndsWith("endif") Then
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                Me.SelectedText = vbLf + vbLf + "EndIf"
                RTBLines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If
        If s.StartsWith("for") Then
            If status.StartsWith("error") And status.EndsWith("next") Then
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                Me.SelectedText = vbLf + vbLf + "Next"
                RTBLines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If
        If s.StartsWith("select") Then
            If status.StartsWith("warning") And status.EndsWith("endselect") Then
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                Me.SelectedText = vbLf + "Case " + vbLf + "EndSelect"
                RTBLines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If

        If s.StartsWith("label") Then
            If (status.StartsWith("ok: this label") And Not s.EndsWith("eventstop") And Not s.EndsWith("eventtimer")) Or _
                status.StartsWith("warning") Or status.StartsWith("error") Then
                ' ---------------------------------------- do not complete if the following line is used
                If FirstSelectedLine + 1 < RTBLines.Length AndAlso _
                   RTBLines(FirstSelectedLine + 1).Trim <> "" Then
                    Return False
                End If
                ' --------------------------------------------------------------------------------------
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                If s.EndsWith("eventstop") Then
                    Me.SelectedText = vbLf + vbLf + "End"
                Else
                    Me.SelectedText = vbLf + vbLf + "Return"
                End If
                RTBLines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If
    End Function

    Private Sub SetAutoCompletionCursorPosition()
        If AutoCompletion_FinalCursorPosition > 0 Then
            Me.SelectionStart = AutoCompletion_FinalCursorPosition
            AutoCompletion_FinalCursorPosition = 0
        End If
    End Sub


    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '   UNDO and HISTORY
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################

    Friend Sub UndoAndHistory_AddEntry()
        ' ----------------------------------------------------------------
        If Not PaintEventEnabled Then Exit Sub
        ' ----------------------------------------------------------------
        Undo_AddEntry()
        History_AddEntry()
        ' ----------------------------------------------------------------
        UndoAndHistoryTextReady = False
    End Sub

    ' ===============================================================================================
    '  PRIVATE UNDO FUNCTIONS
    ' ===============================================================================================
    Private Structure UndoEntry
        Dim ActionName As String
        Dim RTF_Text As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
    End Structure
    Private UndoBuffer As List(Of UndoEntry) = New List(Of UndoEntry)
    Private Undo_Index As Int32
    Friend Sub Undo_Clear()
        UndoBuffer.Clear()
        Undo_Index = 0
        UndoAndHistory_AddEntry()
        TextModified = False
    End Sub
    Private Sub Undo_AddEntry()
        Dim newentry As UndoEntry
        newentry.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = Me.SelectionStart
        newentry.SelectionLength = Me.SelectionLength
        newentry.ActionName = ""
        newentry.RTF_Text = Me.Rtf
        ' ----------------------------------------------------------------
        While UndoBuffer.Count > Undo_Index
            UndoBuffer.RemoveAt(UndoBuffer.Count - 1)
        End While
        ' ----------------------------------------------------------------
        UndoBuffer.Add(newentry)
        Undo_Index += 1
        'Debug.Print(UndoBuffer.Count.ToString)
        ' ----------------------------------------------------------------
        TextModified = True
        TextMustBeSaved = True
        TextMustBeIndented = True
        ' ---------------------------------------------------------------- reinit user variables
        ReinitUserVariables()
    End Sub

    ' ===============================================================================================
    '  UNDO and REDO commands
    ' ===============================================================================================
    Friend Sub ExecUndo()
        ' ---------------------------------------------------------------- 
        TestKeywords_PrepareKeywordsToRecolor()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- undo the RTF text
        Undo_Index -= 1
        If Undo_Index < 1 Then Undo_Index = 1
        Dim undo As UndoEntry
        undo = UndoBuffer(Undo_Index - 1)
        Me.Rtf = undo.RTF_Text
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(undo.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(undo.SelectionStart, 0)
        'Me.Select(undo.SelectionStart, undo.SelectionLength)
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- mark as not-modified
        If Undo_Index < 2 Then TextModified = False
        ' ---------------------------------------------------------------- reinit user variables
        RTBLines = Me.Lines
        ReinitUserVariables()
        ' ---------------------------------------------------------------- 
        ShowSuggestions()
        ShowLinesCount()
        ' ---------------------------------------------------------------- 
        TestKeywords_RecolorLines()
        ColorAllLines_StartingWithVariable()
    End Sub
    Friend Sub ExecRedo()
        ' ---------------------------------------------------------------- 
        TestKeywords_PrepareKeywordsToRecolor()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- undo the RTF text
        If Undo_Index > UndoBuffer.Count - 1 Then Undo_Index = UndoBuffer.Count - 1
        Dim undo As UndoEntry = UndoBuffer(Undo_Index)
        Me.Rtf = undo.RTF_Text
        Undo_Index += 1
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(undo.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(undo.SelectionStart, 0)
        'Me.Select(undo.SelectionStart, undo.SelectionLength)
        ' ---------------------------------------------------------------- 
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- mark as modified
        TextModified = True
        TextMustBeSaved = True
        TextMustBeIndented = True
        ' ---------------------------------------------------------------- reinit user variables
        RTBLines = Me.Lines
        ReinitUserVariables()
        ' ---------------------------------------------------------------- 
        ShowSuggestions()
        ShowLinesCount()
        ' ---------------------------------------------------------------- 
        TestKeywords_RecolorLines()
        ColorAllLines_StartingWithVariable()
    End Sub
    Friend Function Editor_CanUndo() As Boolean
        Return Undo_Index >= 2
    End Function
    Friend Function Editor_CanRedo() As Boolean
        Return Undo_Index <= UndoBuffer.Count - 1
    End Function


    ' ===============================================================================================
    '   PRIVATE HISTORY FUNCTIONS
    ' ===============================================================================================
    Private Structure HistoryEntry
        Dim ActionName As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
    End Structure
    Private HistoryBuffer As List(Of HistoryEntry) = New List(Of HistoryEntry)
    Private History_Index As Int32
    Friend Sub History_Clear()
        HistoryBuffer.Clear()
        History_Index = 0
        History_AddEntry()
    End Sub
    Friend Sub History_AddEntry()
        If Not PaintEventEnabled Then Exit Sub
        ' ----------------------------------------------------------------
        Dim newentry As HistoryEntry
        newentry.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = Me.SelectionStart
        newentry.SelectionLength = Me.SelectionLength
        newentry.ActionName = ""
        ' ----------------------------------------------------------------
        'While HistoryBuffer.Count > History_Index
        '    HistoryBuffer.RemoveAt(HistoryBuffer.Count - 1)
        'End While
        ' ----------------------------------------------------------------
        HistoryBuffer.Add(newentry)
        History_Index += 1
    End Sub

    ' ===============================================================================================
    '  HISTORY commands
    ' ===============================================================================================
    Friend Sub History_Back()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- back
        History_Index -= 1
        If History_Index < 1 Then History_Index = 1
        Dim hist As HistoryEntry
        hist = HistoryBuffer(History_Index - 1)
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(hist.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(hist.SelectionStart, hist.SelectionLength)
        ' ----------------------------------------------------------------
        Me.ShowSuggestions()
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- 
        ShowLinesCount()
    End Sub
    Friend Sub History_Forward()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- forward
        If History_Index > HistoryBuffer.Count - 1 Then History_Index = HistoryBuffer.Count - 1
        Dim hist As HistoryEntry = HistoryBuffer(History_Index)
        History_Index += 1
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(hist.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(hist.SelectionStart, hist.SelectionLength)
        ' ----------------------------------------------------------------
        Me.ShowSuggestions()
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ----------------------------------------------------------------
        ShowLinesCount()
    End Sub
    Friend Function Editor_CanHistoryBack() As Boolean
        Return History_Index >= 2
    End Function
    Friend Function Editor_CanHistoryForward() As Boolean
        Return History_Index <= HistoryBuffer.Count - 1
    End Function



    ' ###############################################################################################
    ' ###############################################################################################
    ' ===============================================================================================
    '   EVENTS
    ' ===============================================================================================
    ' ###############################################################################################
    ' ###############################################################################################


    ' ===============================================================================================
    '   KEY DOWN / KEY UP
    ' ===============================================================================================
    Private OverwriteMode As Boolean = False
    Private LastKeyPressed As Keys

    '  UNDO - Undo entries are saved only after important text changes
    Private UndoAndHistoryTextReady As Boolean = False

    Private Sub SyntaxRichTextBox_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        '
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete, Keys.Back
            Case Keys.Left, Keys.Right, Keys.Up, Keys.Down, Keys.PageUp, Keys.PageDown, Keys.Home, Keys.End
            Case Else
                UndoAndHistoryTextReady = True
        End Select
        '
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete, Keys.Back
                If UndoAndHistoryTextReady Then UndoAndHistory_AddEntry()
                TextMustBeIndented = True
            Case Keys.Tab
                If SelectedText.Length > 0 Then
                    IndentSelectedLines(If(e.Shift, -1, 1))
                    e.Handled = True
                    e.SuppressKeyPress = True
                End If
            Case Keys.Insert
                OverwriteMode_Toggle(Me)
        End Select
        ' -------------------------------------------------- 
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Up, Keys.Down
                TrimRightSpacesFromLine(Me.GetLineFromCharIndex(Me.SelectionStart))
        End Select
        ' -------------------------------------------------- 
        LastKeyPressed = e.KeyCode
        RTBLines = Me.Lines()
        ' -------------------------------------------------- auto completion
        If e.KeyCode = Keys.Enter Then
            If TestAutoCompletion() Then
                e.Handled = True
            End If
        End If
        ' -------------------------------------------------- test keywords and eventually recolor
        TestKeywords_PrepareKeywordsToRecolor()
        ' -------------------------------------------------- suppress RTB sounds
        RichTextBox_SuppressSounds(sender, e)
    End Sub

    Private Sub SyntaxRichTextBox_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        '
        ' -------------------------------------------------- THIS ENABLES MULTIPAGE SELECTION WITH SHIFT
        If e.KeyCode = Keys.ShiftKey Then Return
        '
        ' -------------------------------------------------- THIS ENABLES MULTIPAGE CTRL-C and CTRL-A
        '                                                                     and also SHIFT + ARROWS
        Static LastKeyWasControlOrShift_With_C_A_Arrows As Boolean = False
        If LastKeyWasControlOrShift_With_C_A_Arrows Then
            LastKeyWasControlOrShift_With_C_A_Arrows = False
            If e.KeyCode = Keys.ControlKey Or e.KeyCode = Keys.ShiftKey Then
                Return
            End If
        End If
        If (e.KeyCode = Keys.C Or e.KeyCode = Keys.A Or _
            e.KeyCode = Keys.Up Or e.KeyCode = Keys.Down Or _
            e.KeyCode = Keys.PageUp Or e.KeyCode = Keys.PageDown Or _
            e.KeyCode = Keys.Left Or e.KeyCode = Keys.Right) AndAlso (e.Control Or e.Shift) Then
            LastKeyWasControlOrShift_With_C_A_Arrows = True
            Return
        End If
        '
        ' -------------------------------------------------- 
        RTBLines = Me.Lines()
        ' -------------------------------------------------- REINIT USER VARIABLES
        Dim SelectedLine As Int32 = GetFirstSelectedLine()
        If RTBLines.Length > 0 AndAlso RTBLines(SelectedLine).ToLower.StartsWith("variable") Then
            ReinitUserVariables()
        End If
        ' -------------------------------------------------- finish the autocompletion
        SetAutoCompletionCursorPosition()
        ' -------------------------------------------------- TEST KEYWORDS AND EVENTUALLY RECOLOR LINES
        If e.KeyCode = Keys.Left Or e.KeyCode = Keys.Right Then
            ' Exclude arrows to SpeedUp Right-Left movements 
        Else
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
        End If
        ' -------------------------------------------------- correct the "already used var" erroneous suggestion
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete, Keys.Back
                ShowSuggestions()
        End Select
        ' -------------------------------------------------- show the Overwrite Caret
        OverwriteMode_ShowCaret(Me)
    End Sub


    ' ===============================================================================================
    '  DOUBLE CLICK (complete words an toggle breakpoints)
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.DoubleClick
        '
        ' --------------------------------------------------- complete word - QuotedString or single word
        Dim start, length As Int32
        FindQuotedStringPointers(Me.Text, Me.SelectionStart, start, length)
        If length > 0 Then
            SelectionStart = start
            SelectionLength = length
        Else
            FindWordPointers_ShortWords(Me.Text, MouseDownSelectionIndex, start, length)
            SelectionStart = start
            SelectionLength = length
        End If
        ' --------------------------------------------------- toggle breakpoint
        Dim mousepos As Point = PointToClient(MousePosition)
        ToggleBreakPoint(mousepos)
    End Sub


    ' ===============================================================================================
    '   MOUSE DOWN
    ' ===============================================================================================
    Private OldCursorLine As Int32 = -1
    Private MouseDownSelectionIndex As Int32

    Private Sub SyntaxRichTextBox_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        '
        MouseDownSelectionIndex = Me.GetCharIndexFromPosition(e.Location)
        '
        If e.Button = MouseButtons.Right Then
            If Me.SelectionLength = 0 OrElse _
               MouseDownSelectionIndex < Me.SelectionStart OrElse _
               MouseDownSelectionIndex > Me.SelectionStart + Me.SelectionLength Then
                ' --------------------------------------------------- disable
                PaintEventEnabled = False
                Enabled = False
                ' --------------------------------------------------- remove old selection and restore scroll
                Dim firstVisibleLine As Int32 = GetFirstVisibleLine()
                Me.SelectionLength = 0
                ScrollToLine(firstVisibleLine)
                ' --------------------------------------------------- select word
                Me.SelectionStart = MouseDownSelectionIndex
                SelectCompleteWord_Short()
                'SelectCompleteWord_Long()
                ' --------------------------------------------------- enable
                Enabled = True
                PaintEventEnabled = True
            End If
        End If
        OverwriteMode_ShowCaret(Me)
    End Sub

    Private Sub SelectCompleteWord_Short()
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub

    Private Sub SelectCompleteWord_Long()
        Dim start, length As Int32
        FindWordPointers(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub

    ' ===============================================================================================
    '   MOUSE UP
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        ' ---------------------------------------------------- limit the selection to not include LF 
        If SelectedText.Length > 0 Then
            If SelectedText(SelectedText.Length - 1) = vbLf Then
                SelectionLength -= 1
            End If
        End If
    End Sub

    ' ===============================================================================================
    '   SELECTION CHANGED
    ' ===============================================================================================
    Private Sub RTB_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.SelectionChanged
        If Not EventsAreEnabled Then Exit Sub
        If Form1.Running Then Return
        If Not Enabled Then Return
        ' ----------------------------------------------------------------- SUGGESTIONS
        ShowSuggestions()
        ' ----------------------------------------------------------------- AUTO INDENT
        Dim line As Int32 = GetLineFromCharIndex(Me.SelectionStart)
        If line <> OldCursorLine Then
            OldCursorLine = line
            If TextMustBeIndented Then
                ' --------------------------------------------------------- INDENT
                Dim old As Boolean = EventsAreEnabled
                EventsAreEnabled = False
                AutoIndent()
                EventsAreEnabled = old
            End If
        End If
    End Sub

    ' ===============================================================================================
    '   LOST FOCUS
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LostFocus
        If UndoAndHistoryTextReady Then
            UndoAndHistory_AddEntry()
        End If
    End Sub

End Class
